# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from neutron_lib import context as nctx
from neutron_lib.plugins import constants as pconst
from neutron_lib.plugins import directory
import webob.exc

from gbpservice.neutron.services.grouppolicy import config
from gbpservice.neutron.services.servicechain.plugins.ncp import (
    config as sc_cfg)
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_grouppolicy_plugin as test_plugin)


ML2PLUS_PLUGIN = 'gbpservice.neutron.plugins.ml2plus.plugin.Ml2PlusPlugin'
CORE_PLUGIN = ('gbpservice.neutron.tests.unit.services.grouppolicy.'
               'test_resource_mapping.NoL3NatSGTestPlugin')


class CommonNeutronBaseTestCase(test_plugin.GroupPolicyPluginTestBase):

    def setUp(self, policy_drivers=None, core_plugin=None, l3_plugin=None,
              ml2_options=None, sc_plugin=None, qos_plugin=None,
              trunk_plugin=None):
        core_plugin = core_plugin or ML2PLUS_PLUGIN
        policy_drivers = policy_drivers or ['neutron_resources']
        config.cfg.CONF.set_override('policy_drivers',
                                     policy_drivers,
                                     group='group_policy')
        sc_cfg.cfg.CONF.set_override('node_drivers',
                                     ['dummy_driver'],
                                     group='node_composition_plugin')
        sc_cfg.cfg.CONF.set_override('node_plumber', 'dummy_plumber',
                                     group='node_composition_plugin')
        config.cfg.CONF.set_override('allow_overlapping_ips', True)
        super(CommonNeutronBaseTestCase, self).setUp(core_plugin=core_plugin,
                                                     l3_plugin=l3_plugin,
                                                     ml2_options=ml2_options,
                                                     sc_plugin=sc_plugin,
                                                     qos_plugin=qos_plugin,
                                                     trunk_plugin=trunk_plugin)
        self._plugin = directory.get_plugin()
        self._plugin.remove_networks_from_down_agents = mock.Mock()
        self._plugin.is_agent_down = mock.Mock(return_value=False)
        self._context = nctx.get_admin_context()
        self._gbp_plugin = directory.get_plugin(pconst.GROUP_POLICY)
        self._l3_plugin = directory.get_plugin(pconst.L3)
        config.cfg.CONF.set_override('debug', True)

    def get_plugin_context(self):
        return self._plugin, self._context


class TestL2Policy(CommonNeutronBaseTestCase):

    def _test_l2_policy_lifecycle_implicit_l3p(self,
                                               shared=False):
        l2p = self.create_l2_policy(name="l2p1", shared=shared)
        l2p_id = l2p['l2_policy']['id']
        network_id = l2p['l2_policy']['network_id']
        l3p_id = l2p['l2_policy']['l3_policy_id']
        self.assertIsNotNone(network_id)
        self.assertIsNotNone(l3p_id)
        req = self.new_show_request('networks', network_id, fmt=self.fmt)
        res = self.deserialize(self.fmt, req.get_response(self.api))
        self.assertIsNotNone(res['network']['id'])
        self.show_l3_policy(l3p_id, expected_res_status=200)
        self.show_l2_policy(l2p_id, expected_res_status=200)
        self.update_l2_policy(l2p_id, expected_res_status=200,
                              name="new name")
        self.delete_l2_policy(l2p_id, expected_res_status=204)
        self.show_l2_policy(l2p_id, expected_res_status=404)
        req = self.new_show_request('networks', network_id, fmt=self.fmt)
        res = req.get_response(self.api)
        self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)
        self.show_l3_policy(l3p_id, expected_res_status=404)

    def test_unshared_l2_policy_lifecycle_implicit_l3p(self):
        self._test_l2_policy_lifecycle_implicit_l3p()

    def test_shared_l2_policy_lifecycle_implicit_l3p(self):
        self._test_l2_policy_lifecycle_implicit_l3p(shared=True)


class TestL2PolicyRollback(CommonNeutronBaseTestCase):

    def setUp(self, policy_drivers=None,
              core_plugin=None, ml2_options=None, sc_plugin=None):
        policy_drivers = policy_drivers or ['neutron_resources',
                                            'dummy']
        super(TestL2PolicyRollback, self).setUp(policy_drivers=policy_drivers,
                                                core_plugin=core_plugin,
                                                ml2_options=ml2_options,
                                                sc_plugin=sc_plugin)
        self.dummy_driver = directory.get_plugin(
            'GROUP_POLICY').policy_driver_manager.policy_drivers['dummy'].obj

    def test_l2_policy_create_fail(self):
        orig_func = self.dummy_driver.create_l2_policy_precommit
        self.dummy_driver.create_l2_policy_precommit = mock.Mock(
            side_effect=Exception)
        self.create_l2_policy(name="l2p1", expected_res_status=500)
        self.assertEqual([], self._plugin.get_networks(self._context))
        self.assertEqual([], self._gbp_plugin.get_l2_policies(self._context))
        self.assertEqual([], self._gbp_plugin.get_l3_policies(self._context))
        self.dummy_driver.create_l2_policy_precommit = orig_func

    def test_l2_policy_update_fail(self):
        orig_func = self.dummy_driver.update_l2_policy_precommit
        self.dummy_driver.update_l2_policy_precommit = mock.Mock(
            side_effect=Exception)
        l2p = self.create_l2_policy(name="l2p1")
        l2p_id = l2p['l2_policy']['id']
        self.update_l2_policy(l2p_id, expected_res_status=500,
                              name="new name")
        new_l2p = self.show_l2_policy(l2p_id, expected_res_status=200)
        self.assertEqual(l2p['l2_policy']['name'],
                         new_l2p['l2_policy']['name'])
        self.dummy_driver.update_l2_policy_precommit = orig_func

    def test_l2_policy_delete_fail(self):
        orig_func = self.dummy_driver.delete_l2_policy_precommit
        self.dummy_driver.delete_l2_policy_precommit = mock.Mock(
            side_effect=Exception)
        l2p = self.create_l2_policy(name="l2p1")
        l2p_id = l2p['l2_policy']['id']
        network_id = l2p['l2_policy']['network_id']
        l3p_id = l2p['l2_policy']['l3_policy_id']
        self.delete_l2_policy(l2p_id, expected_res_status=500)
        req = self.new_show_request('networks', network_id, fmt=self.fmt)
        res = self.deserialize(self.fmt, req.get_response(self.api))
        self.assertIsNotNone(res['network']['id'])
        self.show_l3_policy(l3p_id, expected_res_status=200)
        self.show_l2_policy(l2p_id, expected_res_status=200)
        self.dummy_driver.delete_l2_policy_precommit = orig_func



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""sc_instance_spec_mapping
"""

# revision identifiers, used by Alembic.
revision = '8e14fcb1587e'
down_revision = '64fa77aca090'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.create_table(
        'sc_instance_spec_mappings',
        sa.Column('servicechain_spec_id',
                  sa.String(length=36),
                  nullable=False),
        sa.Column('servicechain_instance_id',
                  sa.String(length=36),
                  nullable=False),
        sa.ForeignKeyConstraint(['servicechain_spec_id'], ['sc_specs.id']),
        sa.ForeignKeyConstraint(['servicechain_instance_id'],
                                ['sc_instances.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('servicechain_instance_id',
                                'servicechain_spec_id')
    )


def downgrade(active_plugins=None, options=None):
    op.drop_table('sc_instance_spec_mappings')



# Copyright (c) 2019 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from gbpservice._i18n import _


# REVISIT: Auto-PTG is currently config driven to align with the
# config driven behavior of the older driver but is slated for
# removal.
opts = [
    cfg.BoolOpt('create_auto_ptg',
                default=True,
                help=_("Automatically create a PTG when a L2 Policy "
                       "gets created. This is currently an aim_mapping "
                       "policy driver specific feature.")),
    cfg.BoolOpt('create_per_l3p_implicit_contracts',
                default=True,
                help=_("This configuration is set to True to migrate a "
                       "deployment that has l3_policies without implicit "
                       "AIM contracts (these are deployments which have "
                       "AIM implicit contracts per tenant). A Neutron server "
                       "restart is required for this configuration to take "
                       "effect. The creation of the implicit contracts "
                       "happens at the time of the AIM policy driver "
                       "initialization. The configuration can be set to "
                       "False to avoid recreating the implicit contracts "
                       "on subsequent Neutron server restarts. This "
                       "option will be removed in the O release")),
    cfg.BoolOpt('advertise_mtu',  # REVISIT: Move to apic_aim MD.
                default=True,
                help=_('If True, advertise network MTU values if core plugin '
                       'calculates them. MTU is advertised to running '
                       'instances via DHCP and RA MTU options.')),
    cfg.IntOpt('nested_host_vlan',  # REVISIT: Move to apic_aim MD.
               default=4094,
               help=_("This is a locally siginificant VLAN used to provide "
                      "connectivity to the OpenStack VM when configured "
                      "to host the nested domain (Kubernetes/OpenShift).  "
                      "Any traffic originating from the VM and intended "
                      "to go on the Neutron network, is tagged with this "
                      "VLAN. The VLAN is stripped by the Opflex installed "
                      "flows on the integration bridge and the traffic is "
                      "forwarded on the Neutron network.")),
]


cfg.CONF.register_opts(opts, "aim_mapping")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import requests

from neutron.tests import base
from oslo_config import cfg
from oslo_serialization import jsonutils

from gbpservice.contrib.nfp.configurator.drivers.firewall.vyos import (
                                                    vyos_fw_driver as fw_dvr)
from gbpservice.contrib.nfp.configurator.lib import constants as const
from gbpservice.contrib.tests.unit.nfp.configurator.test_data import (
                                                        fw_test_data as fo)


class FwGenericConfigDriverTestCase(base.BaseTestCase):
    """ Implements test cases for driver methods
    of generic config.

    """

    def __init__(self, *args, **kwargs):
        super(FwGenericConfigDriverTestCase, self).__init__(*args, **kwargs)
        self.fo = fo.FakeObjects()
        with mock.patch.object(cfg, 'CONF') as mock_cfg:
            mock_cfg.configure_mock(rest_timeout=120, host='foo')
            self.driver = fw_dvr.FwaasDriver(mock_cfg)
        self.resp = mock.Mock()
        self.fake_resp_dict = {'status': True, 'reason': 'not found!'}
        self.kwargs = self.fo._fake_resource_data()

    def test_configure_static_ip(self):
        """ Implements test case for configure static ip method
        of generic config driver.

        Returns: none

        """

        with mock.patch.object(
                requests, 'post', return_value=self.resp) as mock_post, (
            mock.patch.object(
                self.resp, 'json', return_value=self.fake_resp_dict)), (
            mock.patch.object(
                self.driver, '_configure_log_forwarding',
                return_value=const.STATUS_SUCCESS)):
            self.driver.configure_interfaces(self.fo.context, self.kwargs)

            data = jsonutils.dumps(self.fo.static_ip_data())
            mock_post.assert_called_with(
                self.fo.get_url_for_api('add_static_ip'),
                data=data, headers=self.fo.fake_header,
                timeout=self.fo.timeout)

    def test_configure_interfaces(self):
        """ Implements test case for configure interfaces method
        of generic config driver.

        Returns: none

        """

        with mock.patch.object(
                requests, 'post', return_value=self.resp) as mock_post, (
            mock.patch.object(
                self.resp, 'json', return_value=self.fake_resp_dict)), (
            mock.patch.object(
                self.driver, '_configure_log_forwarding',
                return_value=const.STATUS_SUCCESS)), (
            mock.patch.object(
                self.driver, '_configure_static_ips',
                return_value=const.STATUS_SUCCESS)):
            self.driver.configure_interfaces(self.fo.context, self.kwargs)

            data = jsonutils.dumps(self.fo.data_for_interface)
            mock_post.assert_called_with(self.fo.get_url_for_api('add_inte'),
                                         data=data,
                                         headers=self.fo.fake_header,
                                         timeout=self.fo.timeout)

    def test_clear_interfaces(self):
        """ Implements test case for clear interfaces method
        of generic config driver.

        Returns: none

        """

        self.resp = mock.Mock(status_code=200)
        with mock.patch.object(
                requests, 'delete', return_value=self.resp) as mock_delete, (
            mock.patch.object(
                self.resp, 'json', return_value=self.fake_resp_dict)):
            self.driver.clear_interfaces(self.fo.context, self.kwargs)

            data = jsonutils.dumps(self.fo.data_for_interface)
            mock_delete.assert_called_with(
                                self.fo.get_url_for_api('del_inte'),
                                data=data, headers=self.fo.fake_header,
                                timeout=self.fo.timeout)

    def test_configure_source_routes(self):
        """ Implements test case for configure routes method
        of generic config driver.

        Returns: none

        """

        with mock.patch.object(
                requests, 'post', return_value=self.resp) as mock_post, (
            mock.patch.object(
                self.resp, 'json', return_value=self.fake_resp_dict)):

            self.driver.configure_routes(self.fo.context, self.kwargs)

            data = list()
            data.append(self.fo.data_for_add_src_route[0])
            data.append(self.fo.data_for_add_src_route[1])
            data = jsonutils.dumps(data)
            mock_post.assert_called_with(
                self.fo.get_url_for_api('add_src_route'),
                data=data, headers=self.fo.fake_header,
                timeout=self.fo.timeout)

    def test_delete_source_routes(self):
        """ Implements test case for clear routes method
        of generic config driver.

        Returns: none

        """

        with mock.patch.object(
                requests, 'delete', return_value=self.resp) as mock_delete, (
            mock.patch.object(
                self.resp, 'json', return_value=self.fake_resp_dict)):
            self.driver.clear_routes(
                self.fo.context, self.kwargs)

            data = list()
            data.append(self.fo.data_for_del_src_route[0])
            data.append(self.fo.data_for_del_src_route[1])
            data = jsonutils.dumps(data)
            mock_delete.assert_called_with(
                self.fo.get_url_for_api('del_src_route'),
                data=data, headers=self.fo.fake_header,
                timeout=self.fo.timeout)


class FwaasDriverTestCase(base.BaseTestCase):
    """ Implements test cases for driver methods
    of firewall.

    """

    def __init__(self, *args, **kwargs):
        super(FwaasDriverTestCase, self).__init__(*args, **kwargs)
        self.fo = fo.FakeObjects()
        with mock.patch.object(cfg, 'CONF') as mock_cfg:
            mock_cfg.configure_mock(rest_timeout=self.fo.timeout, host='foo')
            self.driver = fw_dvr.FwaasDriver(mock_cfg)
        self.resp = mock.Mock()
        self.fake_resp_dict = {'status': True,
                               'message': 'something',
                               'config_success': True,
                               'delete_success': True}
        self.fo.firewall = self.fo._fake_firewall_obj()
        self.firewall = jsonutils.dumps(self.fo.firewall)

    def test_create_firewall_fwaasdriver(self):
        """ Implements test case for create firewall method
        of firewall's drivers.

        Returns: none

        """

        with mock.patch.object(
                requests, 'post', return_value=self.resp) as mock_post, (
            mock.patch.object(
                self.resp, 'json', return_value=self.fake_resp_dict)):
            mock_post.configure_mock(status_code=200)
            self.driver.create_firewall(self.fo.firewall_api_context(),
                                        self.fo.firewall, self.fo.host)
            mock_post.assert_called_with(self.fo.get_url_for_api('config_fw'),
                                         data=self.firewall,
                                         headers=self.fo.fake_header,
                                         timeout=self.fo.timeout)

    def test_update_firewall_fwaasdriver(self):
        """ Implements test case for update firewall method
        of firewall's drivers.

        Returns: none

        """

        with mock.patch.object(
                requests, 'put', return_value=self.resp) as mock_put, (
            mock.patch.object(
                self.resp, 'json', return_value=self.fake_resp_dict)):
            self.driver.update_firewall(self.fo.firewall_api_context(),
                                        self.fo.firewall, self.fo.host)
            mock_put.assert_called_with(self.fo.get_url_for_api('update_fw'),
                                        data=self.firewall,
                                        headers=self.fo.fake_header,
                                        timeout=self.fo.timeout)

    def test_delete_firewall_fwaasdriver(self):
        """ Implements test case for delete firewall method
        of firewall's drivers.

        Returns: none

        """

        with mock.patch.object(
                requests, 'delete', return_value=self.resp) as mock_delete, (
            mock.patch.object(
                self.resp, 'json', return_value=self.fake_resp_dict)):
            self.driver.delete_firewall(self.fo.firewall_api_context(),
                                        self.fo.firewall, self.fo.host)
            mock_delete.assert_called_with(
                self.fo.get_url_for_api('delete_fw'),
                data=self.firewall, headers=self.fo.fake_header,
                timeout=self.fo.timeout)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import collections
import multiprocessing
import operator
import os
import pickle
import sys
import time
import zlib

import eventlet
eventlet.monkey_patch()

from oslo_config import cfg as oslo_config
from oslo_service import service as oslo_service
import six

from gbpservice.nfp.core import cfg as nfp_cfg
from gbpservice.nfp.core import common as nfp_common
from gbpservice.nfp.core import context
from gbpservice.nfp.core import event as nfp_event
from gbpservice.nfp.core import launcher as nfp_launcher
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.core import manager as nfp_manager
from gbpservice.nfp.core import rpc as nfp_rpc
from gbpservice.nfp.core import worker as nfp_worker

# REVISIT (mak): Unused, but needed for orchestrator,
# remove from here and add in orchestrator
from neutron.common import config

LOG = nfp_logging.getLogger(__name__)
PIPE = multiprocessing.Pipe
PROCESS = multiprocessing.Process
identify = nfp_common.identify
deque = collections.deque

# REVISIT (mak): fix to pass compliance check
config = config

"""Implements NFP service.

    Base class for nfp modules, modules can invoke methods
    of this class to interact with core.
"""


class NfpService(object):

    def __init__(self, conf):
        self._conf = conf
        self._event_handlers = nfp_event.NfpEventHandlers()
        self._rpc_agents = {}

    def _make_new_event(self, event):
        """Make a new event from the object passed. """
        desc = event.desc
        event_dict = event.__dict__
        event = self.create_event(**event_dict)
        event.desc.from_desc(desc)
        return event

    def get_event_handlers(self):
        return self._event_handlers

    def register_events(self, event_descs, priority=0):
        """Register event handlers with core. """
        nfp_context = context.get()
        module = nfp_context['log_context']['namespace']
        # REVISIT (mak): change name to register_event_handlers() ?
        for event_desc in event_descs:
            self._event_handlers.register(
                event_desc.id, event_desc.handler,
                module, priority=priority)

    def register_rpc_agents(self, agents):
        """Register rpc handlers with core. """

        for agent in agents:
            topic = agent.topic
            try:
                self._rpc_agents[topic]['agents'].append(agent)
            except KeyError:
                self._rpc_agents[topic] = {}
                self._rpc_agents[topic]['agents'] = [agent]

    def new_event(self, **kwargs):
        """Define and return a new event. """
        return self.create_event(**kwargs)

    def create_event(self, **kwargs):
        """To create a new event. """
        event = None
        try:
            event = nfp_event.Event(**kwargs)
        except AssertionError as aerr:
            message = "%s" % (aerr)
            LOG.exception(message)
        return event

    def post_graph(self, graph_nodes, root_node):
        for node in graph_nodes:
            self.post_event(node)

        self.post_event(root_node)

    def post_event(self, event, target=None):
        """Post an event.

            As a base class, it only does the descriptor preparation.
            NfpController class implements the required functionality.
        """
        handler, module = (
            self._event_handlers.get_event_handler(event.id, module=target))
        assert handler, "No handler registered for event %s" % (event.id)
        event.desc.type = nfp_event.SCHEDULE_EVENT
        event.desc.flag = nfp_event.EVENT_NEW
        event.desc.pid = os.getpid()
        event.desc.target = module
        if event.lifetime == -1:
            event.lifetime = nfp_event.EVENT_DEFAULT_LIFETIME
        if not event.context:
            # Log nfp_context for event handling code
            event.context = context.purge()
        event.desc.path_type = event.context['event_desc'].get('path_type')
        event.desc.path_key = event.context['event_desc'].get('path_key')
        return event

    # REVISIT (mak): spacing=0, caller must explicitly specify
    def poll_event(self, event, spacing=2, max_times=sys.maxint):
        """To poll for an event.

            As a base class, it only does the polling
            descriptor preparation.
            NfpController class implements the required functionality.
        """
        nfp_context = context.get()
        module = nfp_context['log_context']['namespace']
        handler, ev_spacing = (
            self._event_handlers.get_poll_handler(event.id, module=module))
        assert handler, "No poll handler found for event %s" % (event.id)
        assert spacing or ev_spacing, "No spacing specified for polling"
        if ev_spacing:
            spacing = ev_spacing
        if event.desc.type != nfp_event.POLL_EVENT:
            event = self._make_new_event(event)
            event.desc.uuid = event.desc.uuid + ":" + "POLL_EVENT"
        event.desc.type = nfp_event.POLL_EVENT
        event.desc.target = module
        event.desc.flag = None

        kwargs = {'spacing': spacing,
                  'max_times': max_times}
        poll_desc = nfp_event.PollDesc(**kwargs)

        setattr(event.desc, 'poll_desc', poll_desc)

        if not event.context:
            # Log nfp_context for event handling code
            event.context = context.purge()
        event.desc.path_type = event.context['event_desc'].get('path_type')
        event.desc.path_key = event.context['event_desc'].get('path_key')
        return event

    def event_complete(self, event, result=None):
        """To declare and event complete. """
        try:
            pickle.dumps(result)
            uuid = event.desc.uuid
            event = self._make_new_event(event)
            event.desc.uuid = uuid
            event.sequence = False
            event.desc.flag = nfp_event.EVENT_COMPLETE
            event.result = result
            event.context = {}
            event.data = {}
            return event
        except Exception as e:
            raise e

    def create_work(self, work):
        """Create a work, collection of events. """
        pass


"""NFP Controller class mixin other nfp classes.

    Nfp modules get the instance of this class when
    they are initialized.
    Nfp modules interact with core using the methods
    of 'Service' class, whose methods are implemented
    in this class.
    Also, it mixes the other nfp core classes to complete
    a nfp module request.
"""


class NfpController(nfp_launcher.NfpLauncher, NfpService):

    def __new__(cls, *args, **kwargs):
        singleton = kwargs.get('singleton', True)
        if singleton is False:
            return object.__new__(cls, *args, **kwargs)

        if not hasattr(cls, '_instance'):
            cls._instance = object.__new__(cls, *args, **kwargs)
        else:
            cls.__init__ = cls.__inited__
        return cls._instance

    def __inited__(self, conf):
        pass

    def __init__(self, conf, singleton=True):
        # Init the super classes.
        nfp_launcher.NfpLauncher.__init__(self, conf)
        NfpService.__init__(self, conf)

        # For book keeping
        self._worker_process = {}
        self._conf = conf
        self._pipe = None
        # Queue to stash events.
        self._stashq = deque()

        self._manager = nfp_manager.NfpResourceManager(conf, self)
        self._worker = nfp_worker.NfpWorker(conf)

        # ID of process handling this controller obj
        self.PROCESS_TYPE = "distributor"

    def compress(self, event):
        # REVISIT (mak) : zip only if length is > than threshold (1k maybe)
        if not event.zipped:
            event.zipped = True
            data = {'context': event.context}
            event.context = {}
            if event.data:
                data['data'] = event.data
            event.data = zlib.compress(str(data))

    def decompress(self, event):
        if event.zipped:
            try:
                data = ast.literal_eval(
                    zlib.decompress(event.data))
                event.data = data.get('data')
                event.context = data['context']
                event.zipped = False
            except Exception as e:
                message = "Failed to decompress event data, Reason: %r" % (
                    e)
                LOG.error(message)
                raise e

    def is_picklable(self, event):
        """To check event is picklable or not.
           For sending event through pipe it must be picklable
        """
        try:
            pickle.dumps(event)
        except Exception as e:
            message = "(event - %s) is not picklable, Reason: %s" % (
                event.identify(), e)
            assert False, message

    def pipe_recv(self, pipe):
        event = None
        try:
            event = pipe.recv()
        except Exception as exc:
            LOG.debug("Failed to receive event from pipe "
                      "with exception - %r - will retry..", (exc))
            eventlet.greenthread.sleep(1.0)
        if event:
            self.decompress(event)
        return event

    def pipe_send(self, pipe, event, resending=False):
        self.is_picklable(event)

        try:
            # If there is no reader yet
            if not pipe.poll():
                self.compress(event)
                pipe.send(event)
                return True
        except Exception as e:
            message = ("Failed to send event - %s via pipe"
                       "- exception - %r - will resend" % (
                            event.identify(), e))
            LOG.debug(message)

        # If the event is being sent by resending task
        # then dont append here, task will put back the
        # event at right location
        if not resending:
            # If couldnt send event.. stash it so that
            # resender task will send event again
            self._stashq.append(event)
            return False

    def _fork(self, args):
        proc = PROCESS(target=self.child, args=args)
        proc.daemon = True
        proc.start()
        return proc

    def _resending_task(self):
        while(True):
            try:
                event = self._stashq.popleft()
                if self.PROCESS_TYPE != "worker":
                    evm = self._manager._get_event_manager(event.desc.worker)
                    LOG.debug("Resending event - %s", (event.identify()))
                    sent = self.pipe_send(evm._pipe, event, resending=True)
                else:
                    sent = self.pipe_send(self._pipe, event, resending=True)
                # Put back in front
                if not sent:
                    self._stashq.appendleft(event)
            except IndexError:
                pass
            except Exception as e:
                message = ("Unexpected exception - %r - while"
                    "sending event - %s" % (e, event.identify()))
                LOG.error(message)

            eventlet.greenthread.sleep(0.1)

    def _manager_task(self):
        while True:
            # Run 'Manager' here to monitor for workers and
            # events.
            self._manager.manager_run()
            eventlet.greenthread.sleep(0.1)

    def _update_manager(self):
        childs = self.get_childrens()
        for pid, wrapper in six.iteritems(childs):
            pipe = wrapper.child_pipe_map[pid]
            # Inform 'Manager' class about the new_child.
            self._manager.new_child(pid, pipe)

    def _process_event(self, event):
        self._manager.process_events([event])

    def get_childrens(self):
        # oslo_process.ProcessLauncher has this dictionary,
        # 'NfpLauncher' derives oslo_service.ProcessLauncher
        return self.children

    def fork_child(self, wrap):
        """Forks a child.

            Creates a full duplex pipe for child & parent
            to communicate.

            Returns: Multiprocess object.
        """

        parent_pipe, child_pipe = PIPE(duplex=True)

        # Registered event handlers of nfp module.
        # Workers need copy of this data to dispatch an
        # event to module.
        proc = self._fork(args=(wrap.service, parent_pipe, child_pipe, self))

        message = ("Forked a new child: %d"
                   "Parent Pipe: % s, Child Pipe: % s") % (
            proc.pid, str(parent_pipe), str(child_pipe))
        LOG.info(message)

        try:
            wrap.child_pipe_map[proc.pid] = parent_pipe
        except AttributeError:
            setattr(wrap, 'child_pipe_map', {})
            wrap.child_pipe_map[proc.pid] = parent_pipe

        self._worker_process[proc.pid] = proc
        return proc.pid

    def launch(self, workers):
        """Launch the controller.

            Uses Oslo Service to launch with configured #of workers.
            Spawns a manager task to manager nfp events & workers.

            :param workers: #of workers to be launched

            Returns: None
        """
        super(NfpController, self).launch_service(
            self._worker, workers=workers)

    def post_launch(self):
        """Post processing after workers launch.

            Tasks which needs to run only on distributor
            process and any other resources which are not
            expected to be forked are initialized here.
        """
        self._update_manager()

        # create and launch rpc service agent for each topic
        for key, value in six.iteritems(self._rpc_agents):
            agents = value['agents']
            # Register NFP RPC managers in priority order,
            # so that on rpc, oslo invokes them in the given order,
            # This is required for NFP where multiple managers of
            # different priority register for same rpc.
            sorted_agents = sorted(
                agents, key=operator.attrgetter('priority'), reverse=True)
            rpc_managers = [agent.manager for agent in sorted_agents]
            service = nfp_rpc.RpcService(topic=key, managers=rpc_managers)
            # Launch rpc_service_agent
            # Use threads for launching service
            launcher = oslo_service.launch(
                self._conf, service, workers=None)

            self._rpc_agents[key]['service'] = service
            self._rpc_agents[key]['launcher'] = launcher

        # One task to manage the resources - workers & events.
        eventlet.spawn_n(self._manager_task)
        eventlet.spawn_n(self._resending_task)
        # Oslo periodic task for state reporting
        nfp_rpc.ReportStateTask(self._conf, self)

    def report_state(self):
        """Invoked by report_task to report states of all agents. """
        for value in self._rpc_agents.values():
            for agent in value['agents']:
                agent.report_state()

    def _verify_graph(self, graph):
        """Checks for sanity of a graph definition.

            Checks if the same node is root node for
            two subgraphs.
            Unwinds graph and return two values -
            graph signature and graph elements.
        """
        graph_sig = {}
        graph_nodes = []
        for parent, childs in six.iteritems(graph):
            puuid = parent.desc.uuid
            assert puuid not in graph_sig.keys(), (
                "Event - %s is already root of subgraph - %s" % (
                    puuid, str(graph_sig[puuid])))
            graph_sig[puuid] = []
            for child in childs:
                graph_sig[puuid].append(child.desc.uuid)
                graph_nodes.append(child)

        return graph_sig, graph_nodes

    def post_graph(self, graph, root, graph_str=''):
        """Post a new graph into the system.

            Graph is definition of events to be
            dispatched in a particular pattern.
        """
        graph_sig, graph_nodes = self._verify_graph(graph)
        graph_data = {
            'id': root.desc.uuid + "_" + graph_str,
            'root': root.desc.uuid,
            'data': graph_sig}

        for graph_node in graph_nodes:
            graph_node.desc.graph = graph_data

        root.desc.graph = graph_data

        super(NfpController, self).post_graph(graph_nodes, root)

    def post_event(self, event, target=None):
        """Post a new event into the system.

            If distributor(main) process posts an event, it
            is delivered to the worker.
            If worker posts an event, it is deliverd to
            distributor for processing, where it can decide
            to loadbalance & sequence events.

            :param event: Object of 'Event' class.

            Returns: None
        """
        event = super(NfpController, self).post_event(event, target=target)
        message = "(event - %s) - New event" % (event.identify())
        LOG.debug(message)
        if self.PROCESS_TYPE == "worker":
            # Event posted in worker context, send it to parent process
            message = ("(event - %s) - new event in worker"
                       "posting to distributor process") % (event.identify())

            LOG.debug(message)
            # Send it to the distributor process
            self.pipe_send(self._pipe, event)
        else:
            message = ("(event - %s) - new event in distributor"
                       "processing event") % (event.identify())
            LOG.debug(message)
            self._manager.process_events([event])

    def poll_event(self, event, spacing=2, max_times=sys.maxint):
        """Post a poll event into the system.

            Core will poll for this event to timeout, after
            timeout registered handler of module is invoked.

            :param event: Object of 'Event' class.
            :param spacing: Spacing at which event should timeout.
            :param max_times: Max #of times the event can timeout,
                after the max_times, event is auto cancelled by
                the core and the registered handler of module
                is invoked.

            Returns: None
        """
        # Poll event can only be posted by worker not by listener process
        if self.PROCESS_TYPE != "worker":
            message = "(event - %s) - poll event in distributor" % (
                event.identify())
            LOG.debug(message)
            # 'Service' class to construct the poll event descriptor
            event = super(NfpController, self).poll_event(
                event, spacing=spacing, max_times=max_times)
            self._manager.process_events([event])
        else:
            '''
            # Only event which is delivered to a worker can be polled for, coz,
            # after event timeouts, it should be delivered to the same worker,
            # hence the check to make sure the correct event is been asked for
            # polling.
            assert event.desc.worker, "No worker for event %s" % (
                event.identify())
            LOG.debug("(event - %s) - poll event in worker" %
                (event.identify()))
            '''
            # 'Service' class to construct the poll event descriptor
            event = super(NfpController, self).poll_event(
                event, spacing=spacing, max_times=max_times)
            # Send to the distributor process.
            self.pipe_send(self._pipe, event)

    def stop_poll_event(self, key, id):
        """To stop the running poll event

        :param key: key of polling event
        :param id: id of polling event
        """
        key = key + ":" + id + ":" + "POLL_EVENT"
        event = self.new_event(id='STOP_POLL_EVENT', data={'key': key})
        event.desc.type = nfp_event.POLL_EVENT
        event.desc.flag = nfp_event.POLL_EVENT_STOP
        if self.PROCESS_TYPE == "worker":
            self.pipe_send(self._pipe, event)
        else:
            self._manager.process_events([event])

    def path_complete_event(self):
        """Create event for path completion
        """
        nfp_context = context.get()
        event = self.new_event(id='PATH_COMPLETE')
        event.desc.path_type = nfp_context['event_desc'].get('path_type')
        event.desc.path_key = nfp_context['event_desc'].get('path_key')
        if self.PROCESS_TYPE == "worker":
            self.pipe_send(self._pipe, event)
        else:
            self._manager.process_events([event])

    def event_complete(self, event, result=None):
        """To mark an event complete.

            Module can invoke this API to mark an event complete.
                a) Next event in sequence will be scheduled.
                b) Event from cache is removed.
                c) Polling for event is stopped.
                d) If the worker dies before event is complete, the
                    event is scheduled to other available workers.

            :param event: Obj of 'Event' class

            Returns: None
        """
        message = "(event - %s) complete" % (event.identify())
        LOG.debug(message)
        event = super(NfpController, self).event_complete(event, result=result)
        if self.PROCESS_TYPE == "distributor":
            self._manager.process_events([event])
        else:
            # Send to the distributor process.
            self.pipe_send(self._pipe, event)


def load_nfp_modules(conf, controller):
    modules_dirs = conf.nfp_modules_path
    pymodules = []
    for _dir in modules_dirs:
        pymodules.extend(load_nfp_modules_from_path(conf, controller,
                                                    _dir))
    return pymodules


def load_nfp_modules_from_path(conf, controller, path):
    """ Load all nfp modules from configured directory. """
    pymodules = []
    nfp_context = context.get()
    try:
        base_module = __import__(path,
                                 globals(), locals(), ['modules'], -1)
        modules_dir = base_module.__path__[0]
        try:
            files = os.listdir(modules_dir)
            for pyfile in set([f for f in files if f.endswith(".py")]):
                try:
                    pymodule = __import__(path,
                                          globals(), locals(),
                                          [pyfile[:-3]], -1)
                    pymodule = eval('pymodule.%s' % (pyfile[:-3]))
                    try:
                        namespace = pyfile[:-3].split(".")[-1]
                        nfp_context['log_context']['namespace'] = namespace
                        pymodule.nfp_module_init(controller, conf)
                        pymodules += [pymodule]
                        message = "(module - %s) - Initialized" % (
                            identify(pymodule))
                        LOG.debug(message)
                    except AttributeError as e:
                        exc_type, exc_value, exc_traceback = sys.exc_info()
                        message = "Traceback: %s" % (exc_traceback)
                        LOG.error(message)
                        message = ("(module - %s) - does not implement"
                                   "nfp_module_init()") % (identify(pymodule))
                        LOG.warning(message)
                except ImportError:
                    message = "Failed to import module %s" % (pyfile)
                    LOG.error(message)
        except OSError:
            message = "Failed to read files from %s" % (modules_dir)
            LOG.error(message)
    except ImportError:
        message = "Failed to import module from path %s" % (
            path)
        LOG.error(message)

    return pymodules


def controller_init(conf, nfp_controller):
    nfp_controller.launch(conf.workers)
    # Wait for conf.workers*1 + 1 secs for workers to comeup
    time.sleep(conf.workers * 1 + 1)
    nfp_controller.post_launch()


def nfp_modules_post_init(conf, nfp_modules, nfp_controller):
    nfp_context = context.get()
    for module in nfp_modules:
        try:
            namespace = module.__name__.split(".")[-1]
            nfp_context['log_context']['namespace'] = namespace
            module.nfp_module_post_init(nfp_controller, conf)
        except AttributeError:
            message = ("(module - %s) - does not implement"
                       "nfp_module_post_init(), ignoring") % (identify(module))
            LOG.debug(message)


def extract_module(args):
    try:
        index = args.index('--module')
        module = args[index + 1]
        args.remove('--module')
        args.remove(module)
        return args, module
    except ValueError:
        print("--module <name> missing from cmd args")
        sys.exit(-1)


def load_module_opts(conf):
    module = conf.module
    # register each opt from <module> section
    # to default section.
    module_opts = eval('conf.%s.keys' % (module))()
    for module_opt in module_opts:
        module_cfg_opt = eval("conf.%s._group._opts['%s']['opt']" % (
            module, module_opt))
        module_cfg_opt_value = eval("conf.%s.%s" % (module, module_opt))
        conf.register_opt(module_cfg_opt)
        conf.set_override(module_opt, module_cfg_opt_value)


def main():
    context.init()
    args, module = extract_module(sys.argv[1:])
    conf = nfp_cfg.init(module, args)
    conf.module = module
    load_module_opts(conf)
    nfp_logging.init_logger(oslo_config.CONF.logger_class)
    nfp_common.init()
    nfp_controller = NfpController(conf)
    # Load all nfp modules from path configured
    nfp_modules = load_nfp_modules(conf, nfp_controller)
    # Init the controller, launch required contexts
    controller_init(conf, nfp_controller)
    # post_init of each module
    nfp_modules_post_init(conf, nfp_modules, nfp_controller)
    # Wait for every exec context to complete
    nfp_controller.wait()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from neutron.tests import base


class BaseTestCase(base.BaseTestCase):
    """ Defines all the dummy resources needed for test_filter.py
    """
    def __init__(self, *args, **kwargs):
        super(BaseTestCase, self).__init__(*args, **kwargs)
        self.service_info = {}

        self.desc1 = ["fip=1.203.1.108",
                      "tunnel_local_cidr=11.0.0.0/24",
                      "user_access_ip=1.203.2.101",
                      "fixed_ip=192.168.0.3", "standby_fip=",
                      "service_vendor=vyos",
                      "stitching_cidr=192.168.0.0/28",
                      "stitching_gateway=192.168.0.1",
                      "mgmt_gw_ip=120.0.0.1"]
        self.desc2 = ["fip=1.203.1.109",
                      "tunnel_local_cidr=12.0.0.0/24",
                      "user_access_ip=1.203.2.102",
                      "fixed_ip=192.168.0.4", "standby_fip=",
                      "service_vendor=vyos",
                      "stitching_cidr=192.168.0.0/28",
                      "stitching_gateway=192.168.1.1",
                      "mgmt_gw_ip=121.0.0.1"]
        self.name = ["aff8163b-f964-4ad7-a222-0e0a6e5593fe-8eacf5cf",
                     "-1e92-4e7b-90c4-cc68ef8c4e88"]

        self.ssl_vpn_connections = [{
                    "tenant_id": "5e67167662f94fd5987e12a68ea6c1d8",
                    "id": "b88b1d77-fbf8-45b5-adc3-9cd5169c7103",
                    "name": "ssl_vpn_connections1",
                    "admin_state_up": True,
                    "status": "ACTIVE",
                    "vpnservice_id": "19d22704-69ea-40c8-8bcf-2e1ffd697e33",
                    "credential_id": "8163b-f964-4ad7-a222-0e0a6e5593feaff",
                    "client_address_pool_cidr": "11.0.0.0/24"
                                     }]

        self.ports = [{
             "status": "ACTIVE",
             "name": "",
             "allowed_address_pairs": [],
             "admin_state_up": True,
             "network_id": "92f423a7-f44e-4726-b453-c8a1369a3ad0",
             "tenant_id": "5e67167662f94fd5987e12a68ea6c1d8",
             "extra_dhcp_opts": [],
             "binding:vnic_type": "normal",
             "device_owner": "network:dhcp",
             "mac_address": "fa:16:3e:01:19:11",
             "fixed_ips": [{
                     "subnet_id": "2670bdcd-1bcf-4b97-858d-ab0d621983cc",
                     "ip_address": "11.0.0.3"
                            },
                           {
                     "subnet_id": "94aee832-935b-4e23-8f90-b6a81b0195b1",
                     "ip_address": "192.168.0.2"
                            }],
             "id": "cfd9fcc0-c27b-478b-985e-8dd73f2c16e8",
             "security_groups": [],
             "device_id": ("dhcpf986c817-fd54-5bae-a8e4-e473b69100d2-"
                           "92f423a7-f44e-4726-b453-c8a1369a3ad0")
                       },
                      {
             "status": "ACTIVE",
             "name": ("aff8163b-f964-4ad7-a222-0e0a6e5593fe-"
                      "ea9ff596-51bc-4381-8aff-ee9f0ef7e319"),
             "allowed_address_pairs": [],
             "admin_state_up": True,
             "network_id": "0ced2567-47a0-4b67-be52-0e9695e8b0e6",
             "tenant_id": "5e67167662f94fd5987e12a68ea6c1d8",
             "extra_dhcp_opts": [],
             "binding:vnic_type": "normal",
             "device_owner": "network:router_interface",
             "mac_address": "fa:16:3e:1b:f2:44",
             "fixed_ips": [{
                    "subnet_id": "ea9ff596-51bc-4381-8aff-ee9f0ef7e319",
                    "ip_address": "11.0.3.2"
                            }],
             "id": "31df0d68-e9ea-4713-a629-29e6d87c2727",
             "security_groups": ["fb44b3f5-a319-4176-9e3b-361c5faafb66"],
             "device_id": "aff8163b-f964-4ad7-a222-0e0a6e5593fe"
                        },
                      {
             "status": "ACTIVE",
             "name": ";".join(self.name),
             "allowed_address_pairs": [],
             "admin_state_up": True,
             "network_id": "2e9652e8-bd95-472a-96b5-6a7939ae0f8d",
             "tenant_id": "5e67167662f94fd5987e12a68ea6c1d8",
             "extra_dhcp_opts": [],
             "binding:vnic_type": "normal",
             "device_owner": "network:router_interface",
             "mac_address": "fa:16:3e:49:44:b3",
             "fixed_ips": [{
                    "subnet_id": "8eacf5cf-1e92-4e7b-90c4-cc68ef8c4e88",
                    "ip_address": "11.0.4.2"
                            }],
             "id": "214eaa12-36c9-45b1-8fee-350ce2ff2dae",
             "security_groups": ["fb44b3f5-a319-4176-9e3b-361c5faafb66"],
             "device_id": "aff8163b-f964-4ad7-a222-0e0a6e5593fe"
                        }]

        self.subnets = [{
                    "name": "apic_owned_ew-consumer",
                    "enable_dhcp": True,
                    "network_id": "0ced2567-47a0-4b67-be52-0e9695e8b0e6",
                    "tenant_id": "5e67167662f94fd5987e12a68ea6c1d8",
                    "dns_nameservers": [],
                    "gateway_ip": "11.0.3.1",
                    "ipv6_ra_mode": None,
                    "allocation_pools": [{"start": "11.0.3.2",
                                          "end": "11.0.3.254"
                                          }],
                    "host_routes": [],
                    "ip_version": 4,
                    "ipv6_address_mode": None,
                    "cidr": "11.0.3.0/24",
                    "id": "ea9ff596-51bc-4381-8aff-ee9f0ef7e319"
                           },
                          {
                    "name": "apic_owned_ew-provider",
                    "enable_dhcp": True,
                    "network_id": "2e9652e8-bd95-472a-96b5-6a7939ae0f8d",
                    "tenant_id": "5e67167662f94fd5987e12a68ea6c1d8",
                    "dns_nameservers": [],
                    "gateway_ip": "11.0.4.1",
                    "ipv6_ra_mode": None,
                    "allocation_pools": [{"start": "11.0.4.2",
                                          "end": "11.0.4.254"
                                          }],
                    "host_routes": [],
                    "ip_version": 4,
                    "ipv6_address_mode": None,
                    "cidr": "11.0.4.0/24",
                    "id": "94aee832-935b-4e23-8f90-b6a81b0195b1"
                            }]

        self.routers = [{
             "status": "ACTIVE",
             "external_gateway_info": {
                        "network_id": (
                            "a413e04d-1431-4b21-8327-d4de25fa604b"),
                        "external_fixed_ips": [{
                                "subnet_id": (
                                    "fcc74b65-dafe-4b74-91fa-028dec8467a8"),
                                "ip_address": "169.254.2.148"
                                                }]},
             "name": "remote-vpn-client-pool-cidr-l3policy-Datacenter-Out",
             "admin_state_up": True,
             "tenant_id": "5e67167662f94fd5987e12a68ea6c1d8",
             "routes": [],
             "id": "61189c93-d8c7-46ff-b1b1-6a6db2b9ae0a"
                         },
                        {
             "status": "ACTIVE",
             "external_gateway_info": {
                        "network_id": (
                            "a413e04d-1431-4b21-8327-d4de25fa604b"),
                        "external_fixed_ips": [{
                                "subnet_id": (
                                    "fcc74b65-dafe-4b74-91fa-028dec8467a8"),
                                "ip_address": "169.254.2.150"
                                                }]},
             "name": "default-Datacenter-Out",
             "admin_state_up": True,
             "tenant_id": "5e67167662f94fd5987e12a68ea6c1d8",
             "routes": [],
             "id": "aff8163b-f964-4ad7-a222-0e0a6e5593fe"
                         }]

        self.vpnservices = [{
             "router_id": "aff8163b-f964-4ad7-a222-0e0a6e5593fe",
             "status": "ACTIVE",
             "name": "VPNService",
             "admin_state_up": True,
             "subnet_id": "94aee832-935b-4e23-8f90-b6a81b0195b1",
             "tenant_id": "5e67167662f94fd5987e12a68ea6c1d8",
             "id": "19d22704-69ea-40c8-8bcf-2e1ffd697e33",
             "description": ";".join(self.desc1)
                             },
                            {
             "router_id": "61189c93-d8c7-46ff-b1b1-6a6db2b9ae0a",
             "status": "ACTIVE",
             "name": "VPNService1",
             "admin_state_up": True,
             "subnet_id": "94aee832-935b-4e23-8f90-b6a81b0195b1",
             "tenant_id": "5e67167662f94fd5987e12a68ea6c1d8",
             "id": "19d22704-69ea-40c8-8bcf-2e1ffd697f44",
             "description": ";".join(self.desc2)
                              }]

        self.ipsecpolicies = [{
                "encapsulation_mode": "tunnel",
                "encryption_algorithm": "3des",
                "pfs": "group5",
                "lifetime":
                {"units": "seconds", "value": 3600},
                "name": "IPsecPolicy",
                "transform_protocol": "esp",
                "tenant_id": "5e67167662f94fd5987e12a68ea6c1d8",
                "id": "b88b1d77-fbf8-45b5-adc3-9cd5169c7102",
                "auth_algorithm": "sha1", "description": ""
                                 }]

        self.ikepolicies = [{
                "encryption_algorithm": "3des",
                "pfs": "group5",
                "name": "IKEPolicy",
                "phase1_negotiation_mode":
                "main", "lifetime":
                {"units": "seconds", "value": 3600},
                "tenant_id": "5e67167662f94fd5987e12a68ea6c1d8",
                "ike_version": "v1",
                "id": "ae1dd05c-ac66-45e5-868e-36f20c9aa222",
                "auth_algorithm": "sha1",
                "description": ""
                            }]

        self.ipsec_site_connections = [{
             "status": "INIT",
             "psk": "sapna",
             "initiator": "bi-directional",
             "name": "site_to_site_connection1",
             "admin_state_up": True,
             "tenant_id": "5e67167662f94fd5987e12a68ea6c1d8",
             "ipsecpolicy_id": "b88b1d77-fbf8-45b5-adc3-9cd5169c7102",
             "auth_mode": "psk",
             "peer_cidrs": ["11.0.1.0/24"],
             "mtu": 1500,
             "ikepolicy_id": "ae1dd05c-ac66-45e5-868e-36f20c9aa222",
             "dpd": {"action": "hold", "interval": 30, "timeout": 120},
             "route_mode": "static",
             "vpnservice_id": "19d22704-69ea-40c8-8bcf-2e1ffd697e33",
             "peer_address": "1.203.2.1",
             "peer_id": "1.203.2.1",
             "id": "9736cb21-4996-4dae-8e66-d13c24c44a8b",
             "description": ";".join(self.desc1)
                                         }]

        # update the below lists as per the future requirements
        self.firewalls = []
        self.firewall_policies = []
        self.firewall_rules = []

    def _test_get_vpn_info(self):
        """Prepares VPN service_info needed for VPN context

        Returns: VPN service info
        """
        self.service_info['vpnservices'] = self.vpnservices
        self.service_info['ikepolicies'] = self.ikepolicies
        self.service_info['ipsecpolicies'] = self.ipsecpolicies
        self.service_info['ipsec_site_conns'] = self.ipsec_site_connections
        self.service_info['ssl_vpn_conns'] = self.ssl_vpn_connections
        self.service_info['routers'] = self.routers
        self.service_info['subnets'] = self.subnets
        return self.service_info

    def _test_get_fw_info(self):
        """Prepares FW service_info needed for FW context

        Returns: FW service info
        """
        self.service_info['firewalls'] = self.firewalls
        self.service_info['firewall_policies'] = self.firewall_policies
        self.service_info['firewall_rules'] = self.firewall_rules
        return self.service_info



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""service_profile
"""

# revision identifiers, used by Alembic.
revision = '9744740aa75c'
down_revision = 'fd98aa15958d'


from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'service_profiles',
        sa.Column('id', sa.String(36), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('name', sa.String(length=50), nullable=True),
        sa.Column('vendor', sa.String(length=50), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('shared', sa.Boolean),
        sa.Column('insertion_mode', sa.String(length=50), nullable=True),
        sa.Column('service_type', sa.String(length=50), nullable=True),
        sa.Column('service_flavor', sa.String(length=1024), nullable=True),
        sa.PrimaryKeyConstraint('id'),
    )

    op.add_column(
        'sc_nodes',
        sa.Column('service_profile_id', sa.String(36), nullable=True)
    )

    op.create_foreign_key('sc_nodes_ibfk_profile', source='sc_nodes',
                          referent='service_profiles',
                          local_cols=['service_profile_id'],
                          remote_cols=['id'])

    op.alter_column("sc_nodes", "service_type",
                    existing_type=sa.String(length=50), nullable=True)


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from oslo_log import log as logging
import six

from gbpservice.neutron.extensions import group_policy
from gbpservice.neutron.services.servicechain.plugins.ncp import exceptions
from gbpservice.neutron.services.servicechain.plugins.ncp import model

TARGET_DESCRIPTION = "%s facing Service Target for node %s in instance %s"
SERVICE_TARGET_NAME_PREFIX = 'service_target_'
LOG = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class NodePlumberBase(object):
    """Node Plumber base Class

    The node plumber is an entity which takes care of plumbing nodes in a
    chain. By node plumbing is intended the creation/disruption of the
    appropriate Neutron and GBP constructs (typically Ports and Policy Targets)
    based on the specific Node needs, taking into account the whole service
    chain in the process. Ideally, this module will ensure that the traffic
    flows as expected according to the user intent.
    """

    @abc.abstractmethod
    def initialize(self):
        """Perform plumber initialization.

        No abstract methods defined below will be called prior to this method
        being called.
        """

    @abc.abstractmethod
    def plug_services(self, context, deployment):
        """Plug services

        Given a deployment, this method is expected to create all the needed
        policy targets / neutron ports placed where needed based on the whole
        chain configuration.
        The expectation is that this method ultimately creates ServiceTarget
        object that will be retrieved by the node drivers at the right time.

        A deployment is a list composed as follows:
        [{'context': node_context,
          'driver': deploying_driver,
          'plumbing_info': node_plumbing_needs},
          ...]
        No assumptions should be made on the order of the nodes as received in
        the deployment, but it can be retrieved by calling
        node_context.current_position
        """

    @abc.abstractmethod
    def unplug_services(self, context, deployment):
        """Plug services

        Given a deployment, this method is expected to destroy all the
        policy targets / neutron ports previously created for this chain
        configuration.
        The expectation is that this method ultimately removes all the
        ServiceTarget related to this particular chain.

        A deployment is a list composed as follows:
        [{'context': node_context,
          'driver': deploying_driver,
          'plumbing_info': node_plumbing_needs},
          ...]
        No assumptions should be made on the order of the nodes as received in
        the deployment, but it can be retrieved by calling
        node_context.current_position
        """

    def _create_service_targets(self, context, part):
        info = part['plumbing_info']
        if not info:
            return
        part_context = part['context']
        provider = part_context.provider
        consumer = part_context.consumer
        management = part_context.management

        self._create_service_target(context, part_context,
                                    info.get('provider', []),
                                    provider, 'provider')
        self._create_service_target(context, part_context,
                                    info.get('consumer', []),
                                    consumer, 'consumer')
        self._create_service_target(context, part_context,
                                    info.get('management', []),
                                    management, 'management')

    def _delete_service_targets(self, context, part):
        part_context = part['context']
        node = part_context.current_node
        instance = part_context.instance
        gbp_plugin = part_context.gbp_plugin
        pts = model.get_service_targets(
            context.session, servicechain_instance_id=instance['id'],
            servicechain_node_id=node['id'])

        for pt in pts:
            try:
                gbp_plugin.delete_policy_target(
                    context.elevated(), pt.policy_target_id)
            except group_policy.PolicyTargetNotFound as ex:
                LOG.debug(ex.message)

    def _create_service_target(self, context, part_context, targets, group,
                               relationship, extra_data=None):
        extra_data = extra_data or {}
        instance = part_context.instance
        node = part_context.current_node
        gbp_plugin = part_context.gbp_plugin
        for target in targets:
            if not group:
                raise exceptions.NotAvailablePTGForTargetRequest(
                    ptg_type=relationship, instance=instance['id'],
                    node=node['id'])
            data = {'policy_target_group_id': group['id'],
                    'description': TARGET_DESCRIPTION % (relationship,
                                                         node['id'],
                                                         instance['id']),
                    'name': SERVICE_TARGET_NAME_PREFIX + '%s_%s_%s' % (
                        relationship, node['id'][:5], instance['id'][:5]),
                    'port_id': None,
                    'cluster_id': ''}
            data.update(extra_data)
            data.update(target)
            pt = gbp_plugin.create_policy_target(context.elevated(),
                                                 {'policy_target': data})
            model.set_service_target(part_context, pt['id'], relationship)

    def _sort_deployment(self, deployment):
        deployment.sort(key=lambda x: x['context'].current_position,
                        reverse=True)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from neutron_lib.api import validators
import six
from sqlalchemy.orm import exc as orm_exc

from gbpservice.common import utils

VALIDATION_PASSED = "passed"
VALIDATION_REPAIRED = "repaired"
VALIDATION_FAILED_REPAIRABLE = "failed repairable"
VALIDATION_FAILED_BINDING_PORTS = "failed binding ports"
VALIDATION_FAILED_UNREPAIRABLE = "failed unrepairable"
VALIDATION_FAILED_WITH_EXCEPTION = "failed with exception"

VALIDATION_RESULT_PRECEDENCE = [
    VALIDATION_PASSED,
    VALIDATION_REPAIRED,
    VALIDATION_FAILED_REPAIRABLE,
    VALIDATION_FAILED_BINDING_PORTS,
    VALIDATION_FAILED_UNREPAIRABLE,
    VALIDATION_FAILED_WITH_EXCEPTION,
]


@six.add_metaclass(abc.ABCMeta)
class PolicyTargetContext(object):
    """Context passed to policy engine for policy_target resource changes.

    A PolicyTargetContext instance wraps a policy_target resource. It provides
    helper methods for accessing other relevant information. Results
    from expensive operations are cached for convenient access.
    """

    @abc.abstractproperty
    def current(self):
        """Return the current state of the policy_target.

        Return the current state of the policy_target, as defined by
        GroupPolicyPlugin.create_policy_target.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the original state of the policy_target.

        Return the original state of the policy_target, prior to a call to
        update_policy_target. Method is only valid within calls to
        update_policy_target_precommit and update_policy_target_postcommit.
        """
        pass

    @abc.abstractmethod
    def set_port_id(self, port_id):
        """Set the port for the policy_target.

        :param port_id: Port to which policy_target is mapped.

        Set the neutron port to which the policy_target is mapped.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class PolicyTargetGroupContext(object):
    """
    Context passed to policy engine for policy_target_group resource changes.

    PolicyTargetContext instance wraps a policy_target_group resource. It
    provides helper methods for accessing other relevant information. Results
    from expensive operations are cached for convenient access.
    """

    @abc.abstractproperty
    def current(self):
        """Return the current state of the policy_target_group.

        Return the current state of the policy_target_group, as defined by
        GroupPolicyPlugin.create_policy_target_group.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the original state of the policy_target_group.

        Return the original state of the policy_target_group, prior to a call
        to update_policy_target_group. Method is only valid within calls to
        update_policy_target_group_precommit and
        update_policy_target_group_postcommit.
        """
        pass

    @abc.abstractmethod
    def set_l2_policy_id(self, l2_policy_id):
        """Set the l2_policy for the policy_target_group.

        :param l2_policy_id: l2_policy for the policy_target_group.

        Set the l2_policy for the policy_target_group.
        """
        pass

    @abc.abstractmethod
    def set_application_policy_group_id(self, application_policy_group_id):
        """Set the application_policy_group for the policy_target_group.

        :param application_policy_group_id: application_policy_group for
        the policy_target_group.

        Set the application_policy_group for the policy_target_group.
        """
        pass

    @abc.abstractmethod
    def set_network_service_policy_id(self, network_service_policy_id):
        """Set the network_service_policy for the policy_target_group.

        :param network_service_policy_id: network_service_policy for the ptg.

        Set the network_service_policy for the policy_target_group.
        """
        pass

    @abc.abstractmethod
    def add_subnet(self, subnet_id):
        """Add the subnet to the policy_target_group.

        :param subnet_id: Subnet to which policy_target_group is mapped.

        Add a neutron subnet to the set of subnets to which the
        policy_target_group is mapped.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class ApplicationPolicyGroupContext(object):
    """Context passed to policy engine for APG resource changes.

    ApplicationPolicyGroupContext instance wraps a application_policy_group
    resource. It provides helper methods for accessing other relevant
    information. Results from expensive operations are cached for convenient
    access.
    """

    @abc.abstractproperty
    def current(self):
        """Return the current state of the application_policy_group.

        Return the current state of the application_policy_group, as defined by
        GroupPolicyPlugin.create_application_policy_group.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the original state of the application_policy_group.

        Return the original state of the application_policy_group, prior to a
        call to update_application_policy_group. Method is only valid within
        calls to update_application_policy_group_precommit and
        update_application_policy_group_postcommit.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class L2PolicyContext(object):
    """Context passed to policy engine for l2_policy resource changes.

    A L2_ContextContext instance wraps an l2_policy resource. It provides
    helper methods for accessing other relevant information. Results
    from expensive operations are cached for convenient access.
    """

    @abc.abstractproperty
    def current(self):
        """Return the current state of the l2_policy.

        Return the current state of the l2_policy, as defined by
        GroupPolicyPlugin.create_l2_policy.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the original state of the l2_policy.

        Return the original state of the l2_policy, prior to a call to
        update_l2_policy. Method is only valid within calls to
        update_l2_policy_precommit and update_l2_policy_postcommit.
        """
        pass

    @abc.abstractmethod
    def set_l3_policy_id(self, l3_policy_id):
        """Set the l3_policy for the l2_policy.

        :param l3_policy_id: l3_policy for the l2_policy.

        Set the l3_policy for the l2_policy.
        """
        pass

    @abc.abstractmethod
    def set_network_id(self, network_id):
        """Set the network for the l2_policy.

        :param network_id: Network to which l2_policy is mapped.

        Set the neutron network to which the l2_policy is mapped.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class L3PolicyContext(object):
    """Context passed to policy engine for l3_policy resource changes.

    A L3PolicyContext instance wraps an l3_policy resource.
    It provides helper methods for accessing other relevant information.
    Results from expensive operations are cached for convenient access.
    """

    @abc.abstractproperty
    def current(self):
        """Return the current state of the l3_policy.

        Return the current state of the l3_policy, as defined by
        GroupPolicyPlugin.create_l3_policy.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the original state of the l3_policy.

        Return the original state of the l3_policy, prior to a call to
        update_l3_policy. Method is only valid within calls to
        update_l3_policy_precommit and update_l3_policy_postcommit.
        """
        pass

    @abc.abstractmethod
    def add_router(self, router_id):
        """Add the router to the l3_policy.

        :param router_id: Router to which l3_policy is mapped.

        Add a neutron router to the set of routers to which the
        l3_policy is mapped.
        """
        pass

    @abc.abstractmethod
    def set_external_fixed_ips(self, external_segment_id, ips):
        """Add the external_fixed_ips to the l3_policy.

        :param external_segment_id: ES to which l3_policy is mapped.
        :param ips: IPs assigned for that ES.
        """
        pass

    @abc.abstractmethod
    def set_external_segment(self, external_segment_id):
        """Add the external_segment to the l3_policy.

        :param external_segment_id: ES to which l3_policy is mapped.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class NetworkServicePolicyContext(object):
    """
    Context passed to policy engine for network_service_policy resource
    changes.

    A NetworkServicePolicyContext instance wraps a network_service_policy
    resource. It provides helper methods for accessing other relevant
    information. Results from expensive operations are cached for convenient
    access.
    """

    @abc.abstractproperty
    def current(self):
        """Return the current state of the network_service_policy.

        Return the current state of the network_service_policy, as defined by
        GroupPolicyPlugin.create_network_service_policy.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the original state of the network_service_policy.

        Return the original state of the network_service_policy, prior to a
        call to
        update_network_service_policy. Method is only valid within calls to
        update_network_service_policy_precommit and
        update_network_service_policy_postcommit.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class PolicyClassifierContext(object):
    """Context passed to policy engine for policy_classifier resource changes.

    An PolicyClassifierContext instance wraps a policy_classifier resource.
    It provides helper methods for accessing other relevant information.
    Results from expensive operations are cached for convenient access.
    """

    @abc.abstractproperty
    def current(self):
        """Return the current state of the policy_classifier.

        Return the current state of the policy_classifier, as defined by
        GroupPolicyPlugin.create_policy_classifier.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the original state of the policy_classifier.

        Return the original state of the policy_classifier, prior to a call to
        update_policy_classifier. Method is only valid within calls to
        update_policy_classifier_precommit and
        update_policy_classifier_postcommit.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class PolicyActionContext(object):
    """Context passed to policy engine for policy_action resource changes.

    An PolicyActionContext instance wraps a policy_action resource.
    It provides helper methods for accessing other relevant information.
    Results from expensive operations are cached for convenient access.
    """

    @abc.abstractproperty
    def current(self):
        """Return the current state of the policy_action.

        Return the current state of the policy_action, as defined by
        GroupPolicyPlugin.create_policy_action.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the original state of the policy_action.

        Return the original state of the policy_action, prior to a call to
        update_policy_action. Method is only valid within calls to
        update_policy_action_precommit and update_policy_action_postcommit.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class PolicyRuleContext(object):
    """Context passed to policy engine for policy_rule resource changes.

    An PolicyRuleContext instance wraps a policy_rule resource.
    It provides helper methods for accessing other relevant information.
    Results from expensive operations are cached for convenient access.
    """

    @abc.abstractproperty
    def current(self):
        """Return the current state of the policy_rule.

        Return the current state of the policy_rule, as defined by
        GroupPolicyPlugin.create_policy_rule.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the original state of the policy_rule.

        Return the original state of the policy_rule, prior to a call to
        update_policy_rule. Method is only valid within calls to
        update_policy_rule_precommit and
        update_policy_rule_postcommit.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class PolicyRuleSetContext(object):
    """
    Context passed to policy engine for changes to policy_rule_set resources.

    PolicyRuleSetContext instance wraps a policy_rule_set resource. It
    provides helper methods for accessing other relevant information. Results
    from expensive operations are cached for convenient access.
    """

    @abc.abstractproperty
    def current(self):
        """Return the current state of the policy_rule_set.

        Return the current state of the policy_rule_set, as defined by
        GroupPolicyPlugin.create_policy_rule_set.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the original state of the policy_rule_set.

        Return the original state of the policy_rule_set, prior to a call to
        update_policy_rule_set. Method is only valid within calls to
        update_policy_rule_set_precommit and update_policy_rule_set_postcommit.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class ExternalSegmentContext(object):
    """Context passed to policy engine for external_segment resource.

    A ExternalSegmentContext instance wraps an external_segment
    resource.
    It provides helper methods for accessing other relevant information.
    Results from expensive operations are cached for convenient access.
    """

    @abc.abstractproperty
    def current(self):
        """Return the current state of the external_segment.

        Return the current state of the external_segment, as defined by
        GroupPolicyPlugin.create_external_segment.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the original state of the external_segment.

        Return the original state of the external_segment, prior to a
        call to update_external_segment. Method is only valid within
        calls to update_external_segment_precommit and
        update_external_segment_postcommit.
        """
        pass

    @abc.abstractmethod
    def add_subnet(self, subnet_id):
        """Add the subnet to the external_segment.

        :param subnet_id: Subnet to which external_segment is mapped.

        Add a neutron subnet to the set of routers to which the
        external_segment is mapped.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class ExternalPolicyContext(object):
    """Context passed to policy engine for external_policy resource.

    A ExternalPolicyContext instance wraps an external_policy
    resource.
    It provides helper methods for accessing other relevant information.
    Results from expensive operations are cached for convenient access.
    """

    @abc.abstractproperty
    def current(self):
        """Return the current state of the external_policy.

        Return the current state of the external_policy, as defined by
        GroupPolicyPlugin.create_external_policy.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the original state of the external_policy.

        Return the original state of the external_policy, prior to a
        call to update_external_policy. Method is only valid within
        calls to update_external_policy_precommit and
        update_external_policy_postcommit.
        """
        pass

    @abc.abstractmethod
    def set_external_segment(self, external_segment_id):
        """Add the external_segment to the external_policy.

        :param external_segment_id: ES to which external_policy is mapped.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class NatPoolContext(object):
    """Context passed to policy engine for nat_pool resource.

    A NatPoolContext instance wraps an nat_pool
    resource.
    It provides helper methods for accessing other relevant information.
    Results from expensive operations are cached for convenient access.
    """

    @abc.abstractproperty
    def current(self):
        """Return the current state of the nat_pool.

        Return the current state of the nat_pool, as defined by
        GroupPolicyPlugin.create_nat_pool.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the original state of the nat_pool.

        Return the original state of the nat_pool, prior to a
        call to update_nat_pool. Method is only valid within
        calls to update_nat_pool_precommit and
        update_nat_pool_postcommit.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class PolicyDriver(object):
    """Define stable abstract interface for Group Policy drivers.

    A policy driver is called on the creation, update, and deletion
    of all Group Policy resources. For every event, there are two methods that
    get called - one within the database transaction (method suffix of
    _precommit), one right afterwards (method suffix of _postcommit).

    Exceptions raised by methods called inside the transaction can
    rollback, but should not make any blocking calls (for example,
    REST requests to an outside controller). Methods called after
    transaction commits can make blocking external calls, though these
    will block the entire process. Exceptions raised in calls after
    the transaction commits may cause the associated resource to be
    deleted.

    Because rollback outside of the transaction is not done in the
    case of update of resources, all data validation must be done within
    methods that are part of the database transaction.
    """

    @abc.abstractmethod
    def initialize(self):
        """Perform driver initialization.

        Called after all drivers have been loaded and the database has
        been initialized. No abstract methods defined below will be
        called prior to this method being called.
        """
        pass

    def create_policy_target_precommit(self, context):
        """Allocate resources for a new policy_target.

        :param context: PolicyTargetContext instance describing the new
        policy_target.
        """
        pass

    def create_policy_target_postcommit(self, context):
        """Create a policy_target.

        :param context: PolicyTargetContext instance describing the new
        policy_target.
        """
        pass

    def update_policy_target_precommit(self, context):
        """Update resources of a policy_target.

        :param context: PolicyTargetContext instance describing the new
        state of the policy_target, as well as the original state prior
        to the update_policy_target call.
        """
        pass

    def update_policy_target_postcommit(self, context):
        """Update a policy_target.

        :param context: PolicyTargetContext instance describing the new
        state of the policy_target, as well as the original state prior
        to the update_policy_target call.
        """
        pass

    def delete_policy_target_precommit(self, context):
        """Delete resources for a policy_target.

        :param context: PolicyTargetContext instance describing the current
        state of the policy_target, prior to the call to delete it.
        """
        pass

    def delete_policy_target_postcommit(self, context):
        """Delete a policy_target.

        :param context: PolicyTargetContext instance describing the current
        state of the policy_target, prior to the call to delete it.
        """
        pass

    def get_policy_target_status(self, context):
        """Get most recent status of a policy_target.

        :param context: PolicyTargetContext instance describing the current
        state of the policy_target, prior to the call to this get. Driver
        can update the status and status_details. This status change will be
        reflected as the new status and status_details of the resource.
        """
        pass

    def create_policy_target_group_precommit(self, context):
        """Allocate resources for a new policy_target_group.

        :param context: PolicyTargetGroupContext instance describing the new
        policy_target_group.
        """
        pass

    def create_policy_target_group_postcommit(self, context):
        """Create a policy_target_group.

        :param context: PolicyTargetGroupContext instance describing the new
        policy_target_group.
        """
        pass

    def update_policy_target_group_precommit(self, context):
        """Update resources of a policy_target_group.

        :param context: PolicyTargetGroupContext instance describing the new
        state of the policy_target_group, as well as the original state prior
        to the update_policy_target_group call.
        """
        pass

    def update_policy_target_group_postcommit(self, context):
        """Update a policy_target_group.

        :param context: PolicyTargetGroupContext instance describing the new
        state of the policy_target_group, as well as the original state prior
        to the update_policy_target_group call.
        """
        pass

    def delete_policy_target_group_precommit(self, context):
        """Delete resources for a policy_target_group.

        :param context: PolicyTargetGroupContext instance describing the
        current state of the policy_target_group, prior to the call to delete
        it.
        """
        pass

    def delete_policy_target_group_postcommit(self, context):
        """Delete a policy_target_group.

        :param context: PolicyTargetGroupContext instance describing the
        current state of the policy_target_group, prior to the call to delete
        it.
        """
        pass

    def get_policy_target_group_status(self, context):
        """Get most recent status of a policy_target_group.

        :param context: PolicyTargetGroupContext instance describing the
        current state of the policy_target_group, prior to the call to this
        get.
        Driver can update the status and status_details. This status change
        will be reflected as the new status and status_details of the resource.
        """
        pass

    def create_application_policy_group_precommit(self, context):
        """Allocate resources for a new application_policy_group.

        :param context: ApplicationPolicyGroupContext instance describing the
        new application_policy_group.
        """
        pass

    def create_application_policy_group_postcommit(self, context):
        """Create a application_policy_group.

        :param context: ApplicationPolicyGroupContext instance describing the
        new application_policy_group.
        """
        pass

    def update_application_policy_group_precommit(self, context):
        """Update resources of a application_policy_group.

        :param context: ApplicationPolicyGroupContext instance describing the
        new state of the application_policy_group, as well as the original
        state prior to the update_application_policy_group call.
        """
        pass

    def update_application_policy_group_postcommit(self, context):
        """Update a application_policy_group.

        :param context: ApplicationPolicyGroupContext instance describing the
        new state of the application_policy_group, as well as the original
        state prior to the update_application_policy_group call.
        """
        pass

    def delete_application_policy_group_precommit(self, context):
        """Delete resources for a application_policy_group.

        :param context: ApplicationPolicyGroupContext instance describing the
        current state of the application_policy_group, prior to the call to
        delete it.
        """
        pass

    def delete_application_policy_group_postcommit(self, context):
        """Delete a application_policy_group.

        :param context: ApplicationPolicyGroupContext instance describing the
        current state of the application_policy_group, prior to the call to
        delete it.
        """
        pass

    def get_application_policy_group_status(self, context):
        """Get most recent status of a application_policy_group.

        :param context: ApplicationPolicyGroupContext instance describing the
        current state of the application_policy_group, prior to the call to
        this get.
        Driver can update the status and status_details. This status change
        will be reflected as the new status and status_details of the resource.
        """
        pass

    def create_l2_policy_precommit(self, context):
        """Allocate resources for a new l2_policy.

        :param context: L2PolicyContext instance describing the new
        l2_policy.
        """
        pass

    def create_l2_policy_postcommit(self, context):
        """Create a l2_policy.

        :param context: L2PolicyContext instance describing the new
        l2_policy.
        """
        pass

    def update_l2_policy_precommit(self, context):
        """Update resources of a l2_policy.

        :param context: L2PolicyContext instance describing the new
        state of the l2_policy, as well as the original state prior
        to the update_l2_policy call.
        """
        pass

    def update_l2_policy_postcommit(self, context):
        """Update a l2_policy.

        :param context: L2PolicyContext instance describing the new
        state of the l2_policy, as well as the original state prior
        to the update_l2_policy call.
        """
        pass

    def delete_l2_policy_precommit(self, context):
        """Delete resources for a l2_policy.

        :param context: L2PolicyContext instance describing the current
        state of the l2_policy, prior to the call to delete it.
        """
        pass

    def delete_l2_policy_postcommit(self, context):
        """Delete a l2_policy.

        :param context: L2PolicyContext instance describing the current
        state of the l2_policy, prior to the call to delete it.
        """
        pass

    def get_l2_policy_status(self, context):
        """Get most recent status of a l2_policy.

        :param context: L2PolicyContext instance describing the current
        state of the l2_policy, prior to the call to this get.
        Driver can update the status and status_details. This status change
        will be reflected as the new status and status_details of the resource.
        """
        pass

    def create_l3_policy_precommit(self, context):
        """Allocate resources for a new l3_policy.

        :param context: L3PolicyContext instance describing the new
        l3_policy.
        """
        pass

    def create_l3_policy_postcommit(self, context):
        """Create a l3_policy.

        :param context: L3PolicyContext instance describing the new
        l3_policy.
        """
        pass

    def update_l3_policy_precommit(self, context):
        """Update resources of a l3_policy.

        :param context: L3PolicyContext instance describing the new
        state of the l3_policy, as well as the original state prior
        to the update_l3_policy call.
        """
        pass

    def update_l3_policy_postcommit(self, context):
        """Update a l3_policy.

        :param context: L3PolicyContext instance describing the new
        state of the l3_policy, as well as the original state prior
        to the update_l3_policy call.
        """
        pass

    def delete_l3_policy_precommit(self, context):
        """Delete resources for a l3_policy.

        :param context: L3PolicyContext instance describing the current
        state of the l3_policy, prior to the call to delete it.
        """
        pass

    def delete_l3_policy_postcommit(self, context):
        """Delete a l3_policy.

        :param context: L3PolicyContext instance describing the current
        state of the l3_policy, prior to the call to delete it.
        """
        pass

    def get_l3_policy_status(self, context):
        """Get most recent status of a l3_policy.

        :param context: L3PolicyContext instance describing the current
        state of the l3_policy, prior to the call to this get.
        Driver can update the status and status_details. This status change
        will be reflected as the new status and status_details of the resource.
        """
        pass

    def create_policy_classifier_precommit(self, context):
        """Allocate resources for a new policy_classifier.

        :param context: PolicyClassifierContext instance describing the new
        policy_classifier.
        """
        pass

    def create_policy_classifier_postcommit(self, context):
        """Create a policy_classifier.

        :param context: PolicyClassifierContext instance describing the new
        policy_classifier.
        """
        pass

    def update_policy_classifier_precommit(self, context):
        """Update resources of a policy_classifier.

        :param context: PolicyClassifierContext instance describing the new
        state of the policy_classifier, as well as the original state prior
        to the update_policy_classifier call.
        """
        pass

    def update_policy_classifier_postcommit(self, context):
        """Update a policy_classifier.

        :param context: PolicyClassifierContext instance describing the new
        state of the policy_classifier, as well as the original state prior
        to the update_policy_classifier call.
        """
        pass

    def delete_policy_classifier_precommit(self, context):
        """Delete resources for a policy_classifier.

        :param context: PolicyClassifierContext instance describing the current
        state of the policy_classifier, prior to the call to delete it.
        """
        pass

    def delete_policy_classifier_postcommit(self, context):
        """Delete a policy_classifier.

        :param context: PolicyClassifierContext instance describing the current
        state of the policy_classifier, prior to the call to delete it.
        """
        pass

    def get_policy_classifier_status(self, context):
        """Get most recent status of a policy_classifier.

        :param context: PolicyClassifierContext instance describing the current
        state of the policy_classifier, prior to the call to this get.
        Driver can update the status and status_details. This status change
        will be reflected as the new status and status_details of the resource.
        """
        pass

    def create_policy_action_precommit(self, context):
        """Allocate resources for a new policy_action.

        :param context: PolicyActionContext instance describing the new
        policy_action.
        """
        pass

    def create_policy_action_postcommit(self, context):
        """Create a policy_action.

        :param context: PolicyActionContext instance describing the new
        policy_action.
        """
        pass

    def update_policy_action_precommit(self, context):
        """Update resources of a policy_action.

        :param context: PolicyActionContext instance describing the new
        state of the policy_action, as well as the original state prior
        to the update_policy_action call.
        """
        pass

    def update_policy_action_postcommit(self, context):
        """Update a policy_action.

        :param context: PolicyActionContext instance describing the new
        state of the policy_action, as well as the original state prior
        to the update_policy_action call.
        """
        pass

    def delete_policy_action_precommit(self, context):
        """Delete resources for a policy_action.

        :param context: PolicyActionContext instance describing the current
        state of the policy_action, prior to the call to delete it.
        """
        pass

    def delete_policy_action_postcommit(self, context):
        """Delete a policy_action.

        :param context: PolicyActionContext instance describing the current
        state of the policy_action, prior to the call to delete it.
        """
        pass

    def get_policy_action_status(self, context):
        """Get most recent status of a policy_action.

        :param context: PolicyActionContext instance describing the current
        state of the policy_action, prior to the call to this get.
        Driver can update the status and status_details. This status change
        will be reflected as the new status and status_details of the resource.
        """
        pass

    def create_policy_rule_precommit(self, context):
        """Allocate resources for a new policy_rule.

        :param context: PolicyRuleContext instance describing the new
        policy_rule.
        """
        pass

    def create_policy_rule_postcommit(self, context):
        """Create a policy_rule.

        :param context: PolicyRuleContext instance describing the new
        policy_rule.
        """
        pass

    def update_policy_rule_precommit(self, context):
        """Update resources of a policy_rule.

        :param context: PolicyRuleContext instance describing the new
        state of the policy_rule, as well as the original state prior
        to the update_policy_rule call.
        """
        pass

    def update_policy_rule_postcommit(self, context):
        """Update a policy_rule.

        :param context: PolicyRuleContext instance describing the new
        state of the policy_rule, as well as the original state prior
        to the update_policy_rule call.
        """
        pass

    def delete_policy_rule_precommit(self, context):
        """Delete resources for a policy_rule.

        :param context: PolicyRuleContext instance describing the current
        state of the policy_rule, prior to the call to delete it.
        """
        pass

    def delete_policy_rule_postcommit(self, context):
        """Delete a policy_rule.

        :param context: PolicyRuleContext instance describing the current
        state of the policy_rule, prior to the call to delete it.
        """
        pass

    def get_policy_rule_status(self, context):
        """Get most recent status of a policy_rule.

        :param context: PolicyRuleContext instance describing the current
        state of the policy_rule, prior to the call to this get.
        Driver can update the status and status_details. This status change
        will be reflected as the new status and status_details of the resource.
        """
        pass

    def create_policy_rule_set_precommit(self, context):
        """Allocate resources for a new policy_rule_set.

        :param context: PolicyRuleSetContext instance describing the new
        policy_rule_set.
        """
        pass

    def create_policy_rule_set_postcommit(self, context):
        """Create a policy_rule_set.

        :param context: PolicyRuleSetContext instance describing the new
        policy_rule_set.
        """
        pass

    def update_policy_rule_set_precommit(self, context):
        """Update resources of a policy_rule_set.

        :param context: PolicyRuleSetContext instance describing the new
        state of the policy_rule_set, as well as the original state prior
        to the update_policy_rule_set call.
        """
        pass

    def update_policy_rule_set_postcommit(self, context):
        """Update a policy_rule_set.

        :param context: PolicyRuleSetContext instance describing the new
        state of the policy_rule_set, as well as the original state prior
        to the update_policy_rule_set call.
        """
        pass

    def delete_policy_rule_set_precommit(self, context):
        """Delete resources for a policy_rule_set.

        :param context: PolicyRuleSetContext instance describing the current
        state of the policy_rule_set, prior to the call to delete it.
        """
        pass

    def delete_policy_rule_set_postcommit(self, context):
        """Delete a policy_rule_set.

        :param context: PolicyRuleSetContext instance describing the current
        state of the policy_rule_set, prior to the call to delete it.
        """
        pass

    def get_policy_rule_set_status(self, context):
        """Get most recent status of a policy_rule_set.

        :param context: PolicyRuleSetContext instance describing the current
        state of the policy_rule_set, prior to the call to this get.
        Driver can update the status and status_details. This status change
        will be reflected as the new status and status_details of the resource.
        """
        pass

    def create_network_service_policy_precommit(self, context):
        """Allocate resources for a new network service policy.

        :param context: NetworkServicePolicyContext instance describing the new
        network service policy.
        """
        pass

    def create_network_service_policy_postcommit(self, context):
        """Create a network service policy.

        :param context: NetworkServicePolicyContext instance describing the new
        network service policy.
        """
        pass

    def update_network_service_policy_precommit(self, context):
        """Update resources of a network service policy.

        :param context: NetworkServicePolicyContext instance describing the new
        state of the NetworkServicePolicy, as well as the original state prior
        to the update_network_service_policy call.
        """
        pass

    def update_network_service_policy_postcommit(self, context):
        """Update a network service policy.

        :param context: NetworkServicePolicyContext instance describing the new
        state of the NetworkServicePolicy, as well as the original state prior
        to the update_network_service_policy call.
        """
        pass

    def delete_network_service_policy_precommit(self, context):
        """Delete resources for a network service policy.

        :param context: NetworkServicePolicyContext instance describing the
        current state of the NetworkServicePolicy, prior to the call to
        delete it.
        """
        pass

    def delete_network_service_policy_postcommit(self, context):
        """Delete a network service policy.

        :param context: NetworkServicePolicyContext instance describing the
        current state of the NetworkServicePolicy, prior to the call to
        delete it.
        """
        pass

    def get_network_service_policy_status(self, context):
        """Get most recent status of a network_service_policy.

        :param context: NetworkServicePolicyContext instance describing the
        current state of the network_service_policy, prior to the call to this
        get.
        Driver can update the status and status_details. This status change
        will be reflected as the new status and status_details of the resource.
        """
        pass

    def create_external_segment_precommit(self, context):
        """Allocate resources for a new network service policy.

        :param context: ExternalSegmentContext instance describing the
        new network service policy.
        """
        pass

    def create_external_segment_postcommit(self, context):
        """Create a network service policy.

        :param context: ExternalSegmentContext instance describing the
        new network service policy.
        """
        pass

    def update_external_segment_precommit(self, context):
        """Update resources of a network service policy.

        :param context: ExternalSegmentContext instance describing the
        new state of the ExternalSegment, as well as the original state
        prior to the update_external_segment call.
        """
        pass

    def update_external_segment_postcommit(self, context):
        """Update a network service policy.

        :param context: ExternalSegmentContext instance describing the
        new state of the ExternalSegment, as well as the original state
        prior to the update_external_segment call.
        """
        pass

    def delete_external_segment_precommit(self, context):
        """Delete resources for a network service policy.

        :param context: ExternalSegmentContext instance describing the
        current state of the ExternalSegment, prior to the call to
        delete it.
        """
        pass

    def delete_external_segment_postcommit(self, context):
        """Delete a network service policy.

        :param context: ExternalSegmentContext instance describing the
        current state of the ExternalSegment, prior to the call to
        delete it.
        """
        pass

    def get_external_segment_status(self, context):
        """Get most recent status of a external_segment.

        :param context: ExternalSegmentContext instance describing the
        current state of the external_segment, prior to the call to this get.
        Driver can update the status and status_details. This status change
        will be reflected as the new status and status_details of the resource.
        """
        pass

    def create_external_policy_precommit(self, context):
        """Allocate resources for a new network service policy.

        :param context: ExternalPolicyContext instance describing the
        new network service policy.
        """
        pass

    def create_external_policy_postcommit(self, context):
        """Create a network service policy.

        :param context: ExternalPolicyContext instance describing the
        new network service policy.
        """
        pass

    def update_external_policy_precommit(self, context):
        """Update resources of a network service policy.

        :param context: ExternalPolicyContext instance describing the
        new state of the ExternalPolicy, as well as the original state
        prior to the update_external_policy call.
        """
        pass

    def update_external_policy_postcommit(self, context):
        """Update a network service policy.

        :param context: ExternalPolicyContext instance describing the
        new state of the ExternalPolicy, as well as the original state
        prior to the update_external_policy call.
        """
        pass

    def delete_external_policy_precommit(self, context):
        """Delete resources for a network service policy.

        :param context: ExternalPolicyContext instance describing the
        current state of the ExternalPolicy, prior to the call to
        delete it.
        """
        pass

    def delete_external_policy_postcommit(self, context):
        """Delete a network service policy.

        :param context: ExternalPolicyContext instance describing the
        current state of the ExternalPolicy, prior to the call to
        delete it.
        """
        pass

    def get_external_policy_status(self, context):
        """Get most recent status of a external_policy.

        :param context: ExternalPolicyContext instance describing the
        current state of the external_policy, prior to the call to this get.
        Driver can update the status and status_details. This status change
        will be reflected as the new status and status_details of the resource.
        """
        pass

    def create_nat_pool_precommit(self, context):
        """Allocate resources for a new network service policy.

        :param context: NatPoolContext instance describing the
        new network service policy.
        """
        pass

    def create_nat_pool_postcommit(self, context):
        """Create a network service policy.

        :param context: NatPoolContext instance describing the
        new network service policy.
        """
        pass

    def update_nat_pool_precommit(self, context):
        """Update resources of a network service policy.

        :param context: NatPoolContext instance describing the
        new state of the NatPool, as well as the original state
        prior to the update_nat_pool call.
        """
        pass

    def update_nat_pool_postcommit(self, context):
        """Update a network service policy.

        :param context: NatPoolContext instance describing the
        new state of the NatPool, as well as the original state
        prior to the update_nat_pool call.
        """
        pass

    def delete_nat_pool_precommit(self, context):
        """Delete resources for a network service policy.

        :param context: NatPoolContext instance describing the
        current state of the NatPool, prior to the call to
        delete it.
        """
        pass

    def delete_nat_pool_postcommit(self, context):
        """Delete a network service policy.

        :param context: NatPoolContext instance describing the
        current state of the NatPool, prior to the call to
        delete it.
        """
        pass

    def get_nat_pool_status(self, context):
        """Get most recent status of a nat_pool.

        :param context: NatPoolContext instance describing the
        current state of the nat_pool, prior to the call to this get.
        Driver can update the status and status_details. This status change
        will be reflected as the new status and status_details of the resource.
        """
        pass

    # REVISIT(rkukura): Is this needed for all operations, or just for
    # create operations? If its needed for all operations, should the
    # method be specific to the resource and operation, and include
    # the request data (i.e. update_network_pretransaction(self,
    # data))?
    def ensure_tenant(self, plugin_context, tenant_id):
        """Ensure tenant known before creating resource.

        :param plugin_context: Plugin request context.
        :param tenant_id: Tenant owning resource about to be created.

        Called before the start of a transaction creating any new core
        resource, allowing any needed tenant-specific processing to be
        performed.
        """
        pass

    def start_rpc_listeners(self):
        """Start the RPC listeners for the policy drivers.

        When implemented it should return the RPC server object.
        """
        pass

    def validate_state(self, repair):
        """Validate persistent state managed by the driver.

        :param repair: Repair invalid state if True.

        Called from validation tool to validate policy driver's
        persistent state. Returns VALIDATION_PASSED,
        VALIDATION_REPAIRED (only if repair == True),
        VALIDATION_FAILED_REPAIRABLE (only if repair == False),
        VALIDATION_FAILED_UNREPAIRABLE,
        VALIDATION_FAILED_WITH_EXCEPTION, or
        VALIDATION_FAILED_BINDING_PORTS.
        """
        return VALIDATION_PASSED


@six.add_metaclass(abc.ABCMeta)
class ExtensionDriver(object):
    """Define stable abstract interface for Group Policy extension drivers.

    An extension driver extends the core resources implemented by the
    group policy service plugin with additional attributes. Methods
    that process create and update operations for these resources
    validate and persist values for extended attributes supplied
    through the API. Other methods extend the resource dictionaries
    returned from the API operations with the values of the extended
    attributes.
    """

    @abc.abstractmethod
    def initialize(self):
        """Perform driver initialization.

        Called after all drivers have been loaded and the database has
        been initialized. No abstract methods defined below will be
        called prior to this method being called.
        """
        pass

    @abc.abstractproperty
    def extension_alias(self):
        """Supported extension alias.

        Return the alias identifying the Group Policy API extension
        supported by this driver.
        """
        pass

    def process_create_policy_target(self, session, data, result):
        """Process extended attributes for policy_target creation.

        :param session: database session
        :param data: dictionary of incoming policy_target data
        :param result: policy_target dictionary to extend

        Called inside transaction context on session to validate and
        persist any extended policy_target attributes defined by this
        driver. Extended attribute values must also be added to
        result.
        """
        pass

    def process_update_policy_target(self, session, data, result):
        """Process extended attributes for policy_target update.

        :param session: database session
        :param data: dictionary of incoming policy_target data
        :param result: policy_target dictionary to extend

        Called inside transaction context on session to validate and
        update any extended policy_target attributes defined by this
        driver. Extended attribute values, whether updated or not,
        must also be added to result.
        """
        pass

    def extend_policy_target_dict(self, session, result):
        """Add extended attributes to policy_target dictionary.

        :param session: database session
        :param result: policy_target dictionary to extend

        Called inside transaction context on session to add any
        extended attributes defined by this driver to a policy_target
        dictionary to be used for mechanism driver calls and/or
        returned as the result of a policy_target operation.
        """
        pass

    def process_create_policy_target_group(self, session, data, result):
        """Process extended attributes for policy_target_group creation.

        :param session: database session
        :param data: dictionary of incoming policy_target_group data
        :param result: policy_target_group dictionary to extend

        Called inside transaction context on session to validate and
        persist any extended policy_target_group attributes defined by
        this driver. Extended attribute values must also be added to
        result.
        """
        pass

    def process_update_policy_target_group(self, session, data, result):
        """Process extended attributes for policy_target_group update.

        :param session: database session
        :param data: dictionary of incoming policy_target_group data
        :param result: policy_target_group dictionary to extend

        Called inside transaction context on session to validate and
        update any extended policy_target_group attributes defined by
        this driver. Extended attribute values, whether updated or
        not, must also be added to result.
        """
        pass

    def extend_policy_target_group_dict(self, session, result):
        """Add extended attributes to policy_target_group dictionary.

        :param session: database session
        :param result: policy_target_group dictionary to extend

        Called inside transaction context on session to add any
        extended attributes defined by this driver to a
        policy_target_group dictionary to be used for mechanism driver
        calls and/or returned as the result of a policy_target_group
        operation.
        """
        pass

    def process_create_application_policy_group(self, session, data, result):
        """Process extended attributes for application_policy_group creation.

        :param session: database session
        :param data: dictionary of incoming application_policy_group data
        :param result: application_policy_group dictionary to extend

        Called inside transaction context on session to validate and
        persist any extended application_policy_group attributes defined by
        this driver. Extended attribute values must also be added to
        result.
        """
        pass

    def process_update_application_policy_group(self, session, data, result):
        """Process extended attributes for application_policy_group update.

        :param session: database session
        :param data: dictionary of incoming application_policy_group data
        :param result: application_policy_group dictionary to extend

        Called inside transaction context on session to validate and
        update any extended application_policy_group attributes defined by
        this driver. Extended attribute values, whether updated or
        not, must also be added to result.
        """
        pass

    def extend_application_policy_group_dict(self, session, result):
        """Add extended attributes to application_policy_group dictionary.

        :param session: database session
        :param result: application_policy_group dictionary to extend

        Called inside transaction context on session to add any
        extended attributes defined by this driver to a
        application_policy_group dictionary to be used for mechanism driver
        calls and/or returned as the result of a application_policy_group
        operation.
        """
        pass

    def process_create_l2_policy(self, session, data, result):
        """Process extended attributes for l2_policy creation.

        :param session: database session
        :param data: dictionary of incoming l2_policy data
        :param result: l2_policy dictionary to extend

        Called inside transaction context on session to validate and
        persist any extended l2_policy attributes defined by this
        driver. Extended attribute values must also be added to
        result.
        """
        pass

    def process_update_l2_policy(self, session, data, result):
        """Process extended attributes for l2_policy update.

        :param session: database session
        :param data: dictionary of incoming l2_policy data
        :param result: l2_policy dictionary to extend

        Called inside transaction context on session to validate and
        update any extended l2_policy attributes defined by this
        driver. Extended attribute values, whether updated or not,
        must also be added to result.
        """
        pass

    def extend_l2_policy_dict(self, session, result):
        """Add extended attributes to l2_policy dictionary.

        :param session: database session
        :param result: l2_policy dictionary to extend

        Called inside transaction context on session to add any
        extended attributes defined by this driver to a l2_policy
        dictionary to be used for mechanism driver calls and/or
        returned as the result of a l2_policy operation.
        """
        pass

    def process_create_l3_policy(self, session, data, result):
        """Process extended attributes for l3_policy creation.

        :param session: database session
        :param data: dictionary of incoming l3_policy data
        :param result: l3_policy dictionary to extend

        Called inside transaction context on session to validate and
        persist any extended l3_policy attributes defined by this
        driver. Extended attribute values must also be added to
        result.
        """
        pass

    def process_update_l3_policy(self, session, data, result):
        """Process extended attributes for l3_policy update.

        :param session: database session
        :param data: dictionary of incoming l3_policy data
        :param result: l3_policy dictionary to extend

        Called inside transaction context on session to validate and
        update any extended l3_policy attributes defined by this
        driver. Extended attribute values, whether updated or not,
        must also be added to result.
        """
        pass

    def extend_l3_policy_dict(self, session, result):
        """Add extended attributes to l3_policy dictionary.

        :param session: database session
        :param result: l3_policy dictionary to extend

        Called inside transaction context on session to add any
        extended attributes defined by this driver to a l3_policy
        dictionary to be used for mechanism driver calls and/or
        returned as the result of a l3_policy operation.
        """
        pass

    def process_create_policy_classifier(self, session, data, result):
        """Process extended attributes for policy_classifier creation.

        :param session: database session
        :param data: dictionary of incoming policy_classifier data
        :param result: policy_classifier dictionary to extend

        Called inside transaction context on session to validate and
        persist any extended policy_classifier attributes defined by
        this driver. Extended attribute values must also be added to
        result.
        """
        pass

    def process_update_policy_classifier(self, session, data, result):
        """Process extended attributes for policy_classifier update.

        :param session: database session
        :param data: dictionary of incoming policy_classifier data
        :param result: policy_classifier dictionary to extend

        Called inside transaction context on session to validate and
        update any extended policy_classifier attributes defined by
        this driver. Extended attribute values, whether updated or
        not, must also be added to result.
        """
        pass

    def extend_policy_classifier_dict(self, session, result):
        """Add extended attributes to policy_classifier dictionary.

        :param session: database session
        :param result: policy_classifier dictionary to extend

        Called inside transaction context on session to add any
        extended attributes defined by this driver to a
        policy_classifier dictionary to be used for mechanism driver
        calls and/or returned as the result of a policy_classifier
        operation.
        """
        pass

    def process_create_policy_action(self, session, data, result):
        """Process extended attributes for policy_action creation.

        :param session: database session
        :param data: dictionary of incoming policy_action data
        :param result: policy_action dictionary to extend

        Called inside transaction context on session to validate and
        persist any extended policy_action attributes defined by this
        driver. Extended attribute values must also be added to
        result.
        """
        pass

    def process_update_policy_action(self, session, data, result):
        """Process extended attributes for policy_action update.

        :param session: database session
        :param data: dictionary of incoming policy_action data
        :param result: policy_action dictionary to extend

        Called inside transaction context on session to validate and
        update any extended policy_action attributes defined by this
        driver. Extended attribute values, whether updated or not,
        must also be added to result.
        """
        pass

    def extend_policy_action_dict(self, session, result):
        """Add extended attributes to policy_action dictionary.

        :param session: database session
        :param result: policy_action dictionary to extend

        Called inside transaction context on session to add any
        extended attributes defined by this driver to a policy_action
        dictionary to be used for mechanism driver calls and/or
        returned as the result of a policy_action operation.
        """
        pass

    def process_create_policy_rule(self, session, data, result):
        """Process extended attributes for policy_rule creation.

        :param session: database session
        :param data: dictionary of incoming policy_rule data
        :param result: policy_rule dictionary to extend

        Called inside transaction context on session to validate and
        persist any extended policy_rule attributes defined by this
        driver. Extended attribute values must also be added to
        result.
        """
        pass

    def process_update_policy_rule(self, session, data, result):
        """Process extended attributes for policy_rule update.

        :param session: database session
        :param data: dictionary of incoming policy_rule data
        :param result: policy_rule dictionary to extend

        Called inside transaction context on session to validate and
        update any extended policy_rule attributes defined by this
        driver. Extended attribute values, whether updated or not,
        must also be added to result.
        """
        pass

    def extend_policy_rule_dict(self, session, result):
        """Add extended attributes to policy_rule dictionary.

        :param session: database session
        :param result: policy_rule dictionary to extend

        Called inside transaction context on session to add any
        extended attributes defined by this driver to a policy_rule
        dictionary to be used for mechanism driver calls and/or
        returned as the result of a policy_rule operation.
        """
        pass

    def process_create_policy_rule_set(self, session, data, result):
        """Process extended attributes for policy_rule_set creation.

        :param session: database session
        :param data: dictionary of incoming policy_rule_set data
        :param result: policy_rule_set dictionary to extend

        Called inside transaction context on session to validate and
        persist any extended policy_rule_set attributes defined by
        this driver. Extended attribute values must also be added to
        result.
        """
        pass

    def process_update_policy_rule_set(self, session, data, result):
        """Process extended attributes for policy_rule_set update.

        :param session: database session
        :param data: dictionary of incoming policy_rule_set data
        :param result: policy_rule_set dictionary to extend

        Called inside transaction context on session to validate and
        update any extended policy_rule_set attributes defined by this
        driver. Extended attribute values, whether updated or not,
        must also be added to result.
        """
        pass

    def extend_policy_rule_set_dict(self, session, result):
        """Add extended attributes to policy_rule_set dictionary.

        :param session: database session
        :param result: policy_rule_set dictionary to extend

        Called inside transaction context on session to add any
        extended attributes defined by this driver to a
        policy_rule_set dictionary to be used for mechanism driver
        calls and/or returned as the result of a policy_rule_set
        operation.
        """
        pass

    def process_create_network_service_policy(self, session, data, result):
        """Process extended attributes for network_service_policy creation.

        :param session: database session
        :param data: dictionary of incoming network_service_policy data
        :param result: network_service_policy dictionary to extend

        Called inside transaction context on session to validate and
        persist any extended network_service_policy attributes defined
        by this driver. Extended attribute values must also be added
        to result.
        """
        pass

    def process_update_network_service_policy(self, session, data, result):
        """Process extended attributes for network_service_policy update.

        :param session: database session
        :param data: dictionary of incoming network_service_policy data
        :param result: network_service_policy dictionary to extend

        Called inside transaction context on session to validate and
        update any extended network_service_policy attributes defined by this
        driver. Extended attribute values, whether updated or not,
        must also be added to result.
        """
        pass

    def extend_network_service_policy_dict(self, session, result):
        """Add extended attributes to network_service_policy dictionary.

        :param session: database session
        :param result: network_service_policy dictionary to extend

        Called inside transaction context on session to add any
        extended attributes defined by this driver to a
        network_service_policy dictionary to be used for mechanism
        driver calls and/or returned as the result of a
        network_service_policy operation.
        """
        pass

    def process_create_external_segment(self, session, data, result):
        """Process extended attributes for external_segment creation.

        :param session: database session
        :param data: dictionary of incoming external_segment data
        :param result: external_segment dictionary to extend

        Called inside transaction context on session to validate and
        persist any extended external_segment attributes defined
        by this driver. Extended attribute values must also be added
        to result.
        """
        pass

    def process_update_external_segment(self, session, data, result):
        """Process extended attributes for external_segment update.

        :param session: database session
        :param data: dictionary of incoming external_segment data
        :param result: external_segment dictionary to extend

        Called inside transaction context on session to validate and
        update any extended external_segment attributes defined by this
        driver. Extended attribute values, whether updated or not,
        must also be added to result.
        """
        pass

    def extend_external_segment_dict(self, session, result):
        """Add extended attributes to external_segment dictionary.

        :param session: database session
        :param result: external_segment dictionary to extend

        Called inside transaction context on session to add any
        extended attributes defined by this driver to a
        external_segment dictionary to be used for mechanism
        driver calls and/or returned as the result of a
        external_segment operation.
        """
        pass

    def process_create_external_policy(self, session, data, result):
        """Process extended attributes for external_policy creation.

        :param session: database session
        :param data: dictionary of incoming external_policy data
        :param result: external_policy dictionary to extend

        Called inside transaction context on session to validate and
        persist any extended external_policy attributes defined
        by this driver. Extended attribute values must also be added
        to result.
        """
        pass

    def process_update_external_policy(self, session, data, result):
        """Process extended attributes for external_policy update.

        :param session: database session
        :param data: dictionary of incoming external_policy data
        :param result: external_policy dictionary to extend

        Called inside transaction context on session to validate and
        update any extended external_policy attributes defined by this
        driver. Extended attribute values, whether updated or not,
        must also be added to result.
        """
        pass

    def extend_external_policy_dict(self, session, result):
        """Add extended attributes to external_policy dictionary.

        :param session: database session
        :param result: external_policy dictionary to extend

        Called inside transaction context on session to add any
        extended attributes defined by this driver to a
        external_policy dictionary to be used for mechanism
        driver calls and/or returned as the result of a
        external_policy operation.
        """
        pass

    def process_create_nat_pool(self, session, data, result):
        """Process extended attributes for nat_pool creation.

        :param session: database session
        :param data: dictionary of incoming nat_pool data
        :param result: nat_pool dictionary to extend

        Called inside transaction context on session to validate and
        persist any extended nat_pool attributes defined
        by this driver. Extended attribute values must also be added
        to result.
        """
        pass

    def process_update_nat_pool(self, session, data, result):
        """Process extended attributes for nat_pool update.

        :param session: database session
        :param data: dictionary of incoming nat_pool data
        :param result: nat_pool dictionary to extend

        Called inside transaction context on session to validate and
        update any extended nat_pool attributes defined by this
        driver. Extended attribute values, whether updated or not,
        must also be added to result.
        """
        pass

    def extend_nat_pool_dict(self, session, result):
        """Add extended attributes to nat_pool dictionary.

        :param session: database session
        :param result: nat_pool dictionary to extend

        Called inside transaction context on session to add any
        extended attributes defined by this driver to a
        nat_pool dictionary to be used for mechanism
        driver calls and/or returned as the result of a
        nat_pool operation.
        """
        pass

    def _default_process_create(self, session, data, result, type=None,
                                table=None, keys=None):
        """Default process create behavior.

        Gives a default data storing behavior in order to avoid code
        duplication across drivers. Use multiple times to fill multiple
        tables if needed.
        """
        kwargs = dict((x, data[type][x] if
                       validators.is_attr_set(data[type][x]) else None)
                      for x in keys)
        kwargs[type + '_' + 'id'] = result['id']
        record = table(**kwargs)
        session.add(record)
        del kwargs[type + '_' + 'id']
        result.update(kwargs)

    def _default_process_update(self, session, data, result, type=None,
                                table=None, keys=None):
        """Default process update behavior.

        Gives a default data storing behavior in order to avoid code
        duplication across drivers. Use multiple times to fill multiple
        tables if needed.
        """
        try:
            record = (session.query(table).filter_by(**{type + '_' + 'id':
                                                        result['id']}).one())
        except orm_exc.NoResultFound:
            # TODO(ivar) This is a preexisting object. For now just ignore
            # this and return. Each extension driver should be able to specify
            # a default behavior in case this happens.
            return
        for key in keys:
            value = data[type].get(key)
            if validators.is_attr_set(value) and value != getattr(record, key):
                setattr(record, key, value)
            result[key] = getattr(record, key)

    def _default_extend_dict(self, session, result, type=None,
                             table=None, keys=None):
        """Default dictionary extension behavior.

        Gives a default dictionary extension behavior in order to avoid code
        duplication across drivers. Use multiple times to fill from multiple
        tables if needed.
        """
        try:
            record = (session.query(table).filter_by(**{type + '_' + 'id':
                                                        result['id']}).one())
        except orm_exc.NoResultFound:
            # TODO(ivar) This is a preexisting object. For now just ignore
            # this and return. Each extension driver should be able to specify
            # a default behavior in case this happens.
            return
        for key in keys:
            result[key] = getattr(record, key)


def default_extension_behavior(table, keys=None):
    def wrap(func):
        def inner(inst, *args):

            def filter_keys(inst, data, type):
                plural = utils.get_resource_plural(type)
                if keys:
                    return keys
                definition = inst._extension_dict[plural]
                return [x for x in definition if (x in data[type] if data else
                                                  True)]

            name = func.__name__
            if name.startswith('process_create_'):
                # call default process create
                type = name[len('process_create_'):]
                inst._default_process_create(*args, type=type, table=table,
                    keys=filter_keys(inst, args[1], type))
                # Complete result dictionary with unfiltered attributes
                inst._default_extend_dict(args[0], args[2], type=type,
                                          table=table,
                    keys=filter_keys(inst, None, type))
            elif name.startswith('process_update_'):
                # call default process update
                type = name[len('process_update_'):]
                inst._default_process_update(*args, type=type, table=table,
                    keys=filter_keys(inst, args[1], type))

                # Complete result dictionary with unfiltered attributes
                inst._default_extend_dict(args[0], args[2], type=type,
                                          table=table,
                    keys=filter_keys(inst, None, type))
            elif name.startswith('extend_') and name.endswith('_dict'):
                # call default extend dict
                type = name[len('extend_'):-len('_dict')]
                inst._default_extend_dict(*args, type=type, table=table,
                    keys=filter_keys(inst, None, type))
            # Now exec the actual function for postprocessing
            func(inst, *args)
        return inner
    return wrap
